# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''XCUITest API driver
'''

from __future__ import absolute_import, print_function

import base64
import os
import time
import uuid
import re
import json
import subprocess
from six.moves.urllib.request import urlopen
from six.moves.urllib.request import Request

from qt4i.driver.rpc import rpc_method
from qt4i.driver.rpc import RPCEndpoint
from qt4i.driver.xctest.agent import XCUITestAgentManager
from qt4i.driver.xctest.agent import EnumDevice
from qt4i.driver.xctest.webdriverclient.command import Command
from qt4i.driver.xctest.webdriverclient.by import By
from qt4i.driver.tools.dt import DT
from qt4i.driver.tools import mobiledevice
from qt4i.driver.tools.logger import get_logger_path_by_name, clean_expired_log, str_to_time
from qt4i.driver.tools import logger
from qt4i.driver.xctest.webdriverclient.exceptions import NoSuchElementException
from qt4i.driver.xctest.webdriverclient.exceptions import XCTestAgentTimeoutException
from qt4i.driver.util.uimap import UIA_XCT_MAPS
from qt4i.driver.xctest.webdriverclient.errorhandler import ErrorCode
from qt4i.driver.util.modalmap import DeviceProperty
from testbase.conf import settings
from tuia.qpathparser import QPathParser
from pymobiledevice import lockdown


DEFAULT_IP = '127.0.0.1'
DEFAULT_PORT = 8100
TMP_LOAD_FILE = '/tmp/tmpFile'

TMP_DIR_PATH = settings.get('QT4I_TMP_DIR_PATH', '/tmp')

def convert_to_qpath(element):
    '''转换成qpath
    '''
    if 'type' in element:
        element['classname'] = element['type']
        element.pop('type')
    if 'isEnabled' in element:
        element['enabled'] = (element['isEnabled'] == '1')
        element.pop('isEnabled')
    if 'isVisible' in element:
        element['visible'] = (element['isVisible'] == '1')
        element.pop('isVisible')
    if 'rect' in element:
        element['rect']= {
            'origin': {'x': element['rect']['x'], 'y': element['rect']['y']},
            'size'  : {'width': element['rect']['width'], 'height': element['rect']['height']},
        }
    children = []
    for child in element.get('children', []):
        children.append(convert_to_qpath(child))
    element['children'] = children
    return element

class lazy(object): 
    def __init__(self, func): 
        self.func = func 
    
    def __get__(self, instance, cls): 
        val = self.func(instance) 
        setattr(instance, self.func.__name__, val) 
        return val

class Device(RPCEndpoint):
    '''XCUIDevice/XCUIApplication API
    '''
    rpc_name_prefix = "device."
    agent_manager = XCUITestAgentManager()

    def __init__(self, rpc_server, device_id):
        self.rpc_server = rpc_server
        self.udid = device_id
        RPCEndpoint.__init__(self)
    
    @rpc_method   
    def dismiss_alert(self, rule):     
        ''' 弹窗处理
        
        :param rule: 弹窗规则
        :type  rule: list
        ''' 
        env = {
                'rules_of_alert_auto_handle' : rule,
                'flag_alert_auto_handled'    : False
             }
        self.agent.execute(Command.QTA_ALERT_DISMISS, env)      
                  
    @rpc_method
    def stop_all_agents(self):
        '''
        该接口请不要再使用，准备遗弃，请使用host中对应接口
        '''
        XCUITestAgentManager.stop_all_agents()
    
    @lazy
    def agent(self):
        '''延迟初始化agent
        '''
        return self.agent_manager.get_agent(self.udid)

    @rpc_method
    def start_agent(self, agent_ip=DEFAULT_IP, agent_port=DEFAULT_PORT, keep_alive=False, retry=3, timeout=55):
        '''启动Agent，只支持设备主机使用，本地运行不使用该接口
        '''
        self._agent_port = agent_port
        self.agent_manager.start_agent(self.udid, agent_ip, agent_port, keep_alive, retry, timeout)

    @rpc_method
    def restart_agent(self):
        '''重启Agent，只支持设备主机使用，本地运行不使用该接口
        '''
        self.agent_manager.restart_agent(self.udid)

    @rpc_method
    def stop_agent(self):
        '''关闭Agent
        '''
        self.agent_manager.stop_agent(self.udid)
          
    @rpc_method
    def install(self, file_path):
        '''通过udid指定真机安装IPA或模拟器安装APP(注意：真机所用IPA不能在模拟器上安装和使用，IPA与APP相互不兼容)
        
        :param file_path: IPA或APP安装包的路径(注意：当前Mac机必须能访问到该路径)
        :type file_path: str
        :returns: boolean -- True if successfully installed, False if not
        '''
        try :
            dt = DT()
            dt.install(file_path, self.udid)
            return True
        
        except (AttributeError, lockdown.StartServiceError) :
            # 删除 ~/.pymobiledevice中的plist文件          
            dir_user = os.path.expanduser('~')
            dir_plist = os.path.join(dir_user,'.pymobiledevice')
            for root ,dirs ,files in os.walk(dir_plist):  #@UnusedVariable
                for name in files :
                    if name.startswith(self.udid):
                        os.remove(os.path.join(root, name))
                        
            try:
                if not self._isagent_working() :
                    raise RuntimeError('start agent error')
                dt.install(file_path, self.udid)
                return True
            
            except lockdown.NotTrustedError :
                rules = [
                    # 处理信任弹框
                    {
                        'message_text' : '要信任此电脑吗',
                        'button_text'  : '^信任$|^Allow$|^不信任$'
                    },
                ]       
                env = {
                      'rules_of_alert_auto_handle' : rules,
                      'flag_alert_auto_handled'    : True
                    }
                self.agent.execute(Command.QTA_ALERT_DISMISS, env)
                if dt.install(file_path, self.udid) :
                    return True
                raise 
            
        except :
            if dt.install_error == "" :
                import traceback
                dt.install_error = traceback.format_exc()
            raise RuntimeError("install error: %s" % dt.install_error.encode('utf8'))
        
    def _isagent_working(self):
        '''检查当前Agent是否在工作
        '''
        response = self.agent.execute(Command.HEALTH)
        return response['value'] == 'XCTestAgent is ready'

    @rpc_method
    def get_xcode_version(self):
        '''查询Xcode版本
        
        '''
        return DT().get_xcode_version()

    @rpc_method
    def get_ios_version(self):
        '''查询ios版本
        
        '''
        return DT().get_device_by_udid(self.udid)['ios']

    @rpc_method
    def uninstall(self, bundle_id):
        '''通过udid指定设备卸载指定bundle_id的APP
        
        :param bundle_id: APP的bundle_id，例如"tencent.com.sng.test.gn"
        :type bundle_id: str
        :returns: boolean -- if uninstall app successfully, False if not
        '''
        dt = DT()
        if dt.uninstall(bundle_id, self.udid):
            return True
        raise RuntimeError("uninstall error: %s" % dt.uninstall_error.encode('utf8'))
    
    @rpc_method
    def reboot(self):
        '''重启手机
        
        :param device_udid: 设备的udid
        :type device_udid: str
        '''
        return DT().reboot(self.udid)
    
    @rpc_method
    def start_app(self, bundle_id, app_params=None, env=None, trace_template=None, trace_output=None, retry=3, timeout=55):
        '''启动设备上的指定APP
        
        :param bundle_id: 要启动的APP的Bundle ID
        :type bundle_id: str
        :param app_params: APP 使用的参数
        :type app_params: dict
        :param trace_template: 专项测试使用trace_template的路径
        :type trace_template: str
        :param trace_output: 专项测试使用trace_output
        :type trace_output: str
        :param retry: 启动失败重试次数
        :type retry: int
        :param timeout: 单次启动超时 (秒)
        :type timeout: int
        :returns: boolean -- True if start successfully , or raise exception
        ''' 
        desired_caps = {}
        desired_caps['bundleId'] = bundle_id
        ios_version = DT().get_device_by_udid(self.udid)['ios']
        if DT.compare_version(ios_version, '12.0') >= 0:
            app_list = []
            for app in DT().list_apps(self.udid, 'all'):
                app_list.extend(app.keys())
            if bundle_id not in app_list:
                raise Exception("Failed to launch application, %s not exist" % bundle_id)
        if app_params:
            params_list = []
            for k in app_params:
                params_list.append("-%s %s" % (k, app_params[k]))
            desired_caps['arguments'] = params_list
        for _ in range(retry):
            try:
                self.agent.execute(Command.START, desired_caps)
                break
            except XCTestAgentTimeoutException:
                self.restart_agent()
        else:
            raise XCTestAgentTimeoutException('Failed to start app: %s' % bundle_id)       
        if env:
            self.agent.execute(Command.QTA_ALERT_RULES_UPDATE, env)
        return True
    
    @rpc_method
    def stop_app(self, bundle_id):
        '''停止指定app
        
        :returns: boolean -- True if start successfully, False if not
        '''
        return self.agent.execute(Command.QUIT)
    
    @rpc_method
    def get_rect(self):
        '''获取设备屏幕大小
        
        :returns: 坐标和高宽
        '''
        return {
            'origin': {'y': 0, 'x': 0},
            'size'  : self.agent.execute(Command.GET_WINDOW_SIZE)['value'],
        }
    
    @rpc_method
    def get_model(self):
        '''获取设备模型(iPhone/iPad)
        :returns: str
        '''
        if self.agent.has_session():
            return self.agent.capabilities['device']
        else:
            return self.agent.execute(Command.STATUS)['value']['os']['name']
    
    @rpc_method
    def get_name(self):
        '''获取设备名
        
        :returns: str
        '''
        raise NotImplementedError
    
    @rpc_method
    def get_system_version(self):
        '''获取系统版本(9.3.2)
        
        :returns: str
        '''
        return self.agent.execute(Command.STATUS)['value']['os']['version']
    
    @rpc_method
    def capture_screen(self):
        '''截屏
        
        :returns: str -- base64编码后截图数据
        '''
        try:
            base64_img = self.agent.capture_screen()
            if base64_img is None:
                base64_img = self.agent.execute(Command.SCREENSHOT)['value'].encode('ascii')
            else:
                base64_img = base64_img['value']
        except:
            filename = '/tmp/%s.png' % uuid.uuid1()
            if self.agent.type == EnumDevice.Simulator:
                version = DT().get_xcode_version()
                if version >= '8':
                    cmd = ['xcrun', 'simctl', 'io', self.udid,'screenshot',filename]
                    p = subprocess.Popen(cmd,stdin=subprocess.PIPE,stdout=subprocess.PIPE, stderr=subprocess.PIPE,shell=False)
                    p.communicate()
                    retcode = p.poll()
                    if retcode != 0:
                        raise Exception("simulator screenshot error")
                else:
                    raise Exception("simulator screenshot error: Xcode version less than 8 does not support")
            elif not mobiledevice.get_screenshot(self.udid, filename):
                raise Exception("mobiledevice screenshot error")
            with open(filename, "rb") as fd:
                base64_img = base64.b64encode(fd.read())
            os.remove(filename)
        return base64_img

    @rpc_method
    def get_element_tree(self):
        '''从顶层开始获取UI树字典
        
        :returns: dict
        '''
        result = self.agent.execute(Command.GET_ELEMENT_TREE)['value']['tree']
        return convert_to_qpath(result)
    
    @rpc_method
    def get_element_tree_and_capture_screen(self, filepath=None):
        '''获取UI树并截屏(该接口仅限UISpy使用)
        
        :param filepath: 将图片存储至该路径（png格式）
        :type filepath: str
        :returns: dict : { element_tree: {}, capture_screen: file_path }
        '''
        result = {}
        result['element_tree'] = self.get_element_tree()
        result['capture_screen'] = self.capture_screen(filepath)
        return result

    @rpc_method
    def set_alert_auto_handling_rules(self, rules):
        '''设置自动处理Alert规则
        
                    当message_text匹配到Alert框内的任意文本，则点击button_text指定的按钮。
                    文本以段为单元，元素的label、name、value分为三段文本。规则匹配范围是Alert框内的所有元素的文本。
        :param rules: 规则集合
        :type rules: list - [ {"message_text": "message_text", "button_text": "button_text"}, ... ]
        '''
        self.agent.execute(Command.QTA_ALERT_RULES_UPDATE, rules)

    @rpc_method
    def get_alert_auto_handling_rules(self):
        '''获取已设置的自动处理Alert规则
        
        :returns: list - [ {"message_text": "message_text", "button_text": "button_text"}, ... ]
        '''
        raise NotImplementedError

    @rpc_method
    def add_alert_auto_handling_rule(self, message_text, button_text):
        '''自动处理Alert规则，添加一项规则
        
        :param message_text: Alert内文本片段，支持正则表达式（范围是所谓文本，每个元素的label/name/value为文本段）
        :type message_text: str
        :param button_text: Alert内按钮的文本，支持正则表达式（元素的label/name/value）
        :type button_text: str
        '''
        raise NotImplementedError

    @rpc_method
    def clear_alert_auto_handling_rules(self):
        '''自动处理Alert规则，清空所有规则
        '''
        raise NotImplementedError

    @rpc_method
    def disable_alert_auto_handling(self):
        '''关闭自动关闭alert框
        '''
        raise NotImplementedError
    
    @rpc_method
    def enable_alert_auto_handling(self):
        '''打开自动关闭alert框(优先级低于set_alert_auto_handling_rules所设置的策略)
        '''
        raise NotImplementedError
    
    @rpc_method
    def background_app(self, seconds=3):
        '''将App置于后台一定时间
        
        :param seconds: 秒
        :type seconds: int
        '''
        self.agent.execute(Command.DEACTIVATE_APP, {'duration':seconds})
    
    @rpc_method
    def click(self, x, y, options=None):
        '''单击（全局操作）
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        :param options: 屏幕操作扩展参数
        :type options: dict
        '''
        self.agent.execute(Command.QTA_DEVICE_CLICK, {'x':x, 'y':y, 'options':options})['value']
    
    @rpc_method
    def double_click(self, x, y):
        '''双击（全局操作）
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        '''
        self.agent.execute(Command.QTA_DEVICE_DOUBLE_CLICK, {'x':x, 'y':y})['value']
    
    @rpc_method
    def long_click(self, x, y, seconds=3):
        '''长按
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        :param seconds: 秒
        :type seconds: int
        '''
        self.agent.execute(Command.QTA_DEVICE_LONG_CLICK, {'x':x, 'y':y, 'duration':seconds})['value']
    
    @rpc_method
    def drag(self, x0, y0, x1, y1, duration=0.1, repeat=1, interval=1, velocity=1000):
        '''拖拽（全局操作）
        :param x0: 起始横向坐标（从左向右，屏幕百分比）
        :type x0: float
        :param y0: 起始纵向坐标（从上向下，屏幕百分比）
        :type y0: float
        :param x1: 终止横向坐标（从左向右，屏幕百分比）
        :type x1: float
        :param y1: 终止纵向坐标（从上向下，屏幕百分比）
        :type y1: float
        :param duration: 起始坐标按下的时间（秒）
        :type duration: float
        '''
        self.agent.execute(Command.QTA_DEVICE_DRAG, 
                {'x0':x0, 'y0':y0, 'x1':x1, 'y1':y1, 'duration':duration, 
                 'repeat':repeat, 'interval':interval, 'velocity':velocity})['value']
    
    @rpc_method
    def pinch(self, x0, y0, x1, y1, opposing=True, seconds=1):
        '''捏合/展开
        :param x0: 起始横向坐标（从左向右，屏幕百分比）
        :type x0: float
        :param y0: 起始纵向坐标（从上向下，屏幕百分比）
        :type y0: float
        :param x1: 终止横向坐标（从左向右，屏幕百分比）
        :type x1: float
        :param y1: 终止纵向坐标（从上向下，屏幕百分比）
        :type y1: float
        :param opposing: True为展开；False为捏合
        :type opposing: boolean
        :param seconds: 秒
        :type seconds: int
        '''
        raise NotImplementedError
    
    @rpc_method
    def rotate(self, x, y, options=None):
        '''旋转
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        :param options     : 自定义项
        :type options      : dict : {
                            :           duration    : 1,    持续的时间（秒）
                            :           radius      : 50,   半径
                            :           rotation    : 100,  圆周，默认是圆周率PI
                            :           touchCount  : 2     触摸点数量 2 - 5
                            :          }
        '''
        raise NotImplementedError
    
    @rpc_method
    def set_location(self, coordinates, options=None):
        '''设置设备的GPS坐标
        
        :param _coordinates : 经纬度
        :type _coordinates  : dict : {
                            :           latitude    : 100, // 以度为单位的纬度。正值表示纬度赤道以北。负值表示纬度赤道以南。
                            :           longitude   : 100  // 以度为单位的经度。测量是相对于零子午线，用正值向东延伸的经络及经络的负值西侧延伸。
                            :          }
        '''
        raise NotImplementedError
    
    @rpc_method
    def shake(self):
        '''摇晃设备
        '''
        raise NotImplementedError
    
    @rpc_method
    def volume(self, cmd):
        '''调高音量
        '''
        self.agent.execute(Command.QTA_DEVICE_VOLUME, {'value':cmd})
    
    @rpc_method
    def screen_direction(self, direct):
        '''调高音量
        '''
        self.agent.execute(Command.QTA_DEVICE_SCREEN_DIRECTION, {'value':direct})
        
    @rpc_method
    def siri(self, cmd):
        ''' siri交流
        '''
        self.agent.execute(Command.QTA_DEVICE_SIRI, {'value':cmd}) 
    
    @rpc_method
    def unlock(self):
        '''解锁设备
        '''
        self.agent.execute(Command.QTA_DEVICE_UNLOCK)
    
    @rpc_method
    def lock(self, seconds=0):
        '''锁定设备
        
        :param seconds: 秒, 0表示锁定设备；大于0表示锁定一段时间后解锁
        :type seconds: int
        '''
        self.agent.execute(Command.QTA_DEVICE_LOCK)
        
    @rpc_method
    def send_keys(self, keys):
        '''键盘输入
        
        :param keys: 键入的字符串
        :type keys: str
        '''
        self.agent.execute(Command.QTA_DEVICE_SENDKEYS, {'value':keys})['value']
    
    @rpc_method
    def get_driver_log(self, start_time=None):
        '''获取driver日志
        
        :parm start_time 用例执行的开始时间
        :type str
        '''
        driver_log_name = 'xctest_%s' % self.udid
        
        return self._read_to_log(driver_log_name, start_time)

    @rpc_method
    def get_crash_log(self, procname):
        '''通过设备的udid的获取指定进程的最新的crash日志
        
        :param proc_name: app的进程名，可通过xcode查看
        :type proc_name: str        
        :returns: str
        '''
        return DT().get_crash_log(self.udid, procname)
    
    @rpc_method
    def get_log(self, start_time=None):
        '''获取日志信息
        
        :parm start_time 用例执行的开始时间
        :type str
        '''
        
        log_name = 'driverserver_%s' % self.udid
        
        return self._read_to_log(log_name, start_time)
    
    def _read_to_log(self, log_name, start_time):
        '''获取日志内容
        
        :parm end_time 用例结束的时间
        :type time(默认为None，则表示切割的存放日志整个都有效)
        :param log_name 日志名称
        :type str
        :parm start_time 用例执行的开始时间
        :type str
        '''
        #兼容client端为旧版本，server端为新版本，只返回当前文件的日志（日志会过多或过少）
        if not start_time:
            with open(get_logger_path_by_name(log_name)) as fd:
                return fd.read()
        
        log = ''
        log_tmp = ''
        is_log_tmp = '' #当前文件的顶部没时间匹配的待确定是否属于日志内容
        pattern = r'^\[(\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2})'
        flag = False #标记该时间段的日志是否已经取完
        is_flag = False #标记取到的未匹配时间的行是否属于日志内容
        start_time = str_to_time(start_time)
        files = os.listdir(logger.TMP_DIR_PATH)
        
        for f in files:
            log_path = os.path.join(logger.TMP_DIR_PATH, f)
            if not flag and os.path.exists(log_path) and f.find(log_name) != -1:
                with open(log_path, "rb") as fd:
                    for line in fd:
                        line = line.decode("utf-8")
                        match = re.search(pattern, line)
                        if match:
                            cur_time = str_to_time(match.group(1))
                            if cur_time < start_time:
                                flag = True
                                is_log_tmp = ''
                                continue
                            elif cur_time == start_time and is_log_tmp:
                                is_log_tmp = ''
                                log_tmp += line
                            else:
                                is_flag = True
                                log_tmp += line
                        elif is_flag:
                            log_tmp += line
                        else:
                            is_log_tmp += line
                log_tmp = is_log_tmp + log_tmp
                log = log_tmp + log
                log_tmp = ''
        return log
    
    @rpc_method
    def cleanup_log(self):
        '''清理日志
        '''
        clean_expired_log("driverserver_%s" % self.udid)
        clean_expired_log('xctest_%s' % self.udid)
    
    @rpc_method
    def push_file(self, bundle_id, localpath, remotepath):
        '''拷贝Mac本地文件到iOS设备sandbox的指定目录
        
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param localpath: Mac上的文件路径
        :type localpath: str
        :param remotepath: sandbox上的目录，例如：/Library/Caches/test/
        :type remotepath: str        
        :returns: boolean
        '''
        return DT().push_file(self.udid, bundle_id, localpath, remotepath)
    
    @rpc_method
    def download_file_and_push(self, bundle_id, file_url, remotepath):
        '''支持URL上传文件到IOS设备sandbox的指定目录
        
        :param bundle_id: app的bundle id
        :type bundle_id: str
        :parameter file_url 文件下载链接
        :type file_url: str
        :param remotepath: sandbox上的目录，例如：/Library/Caches/test/
        :type remotepath: str
        :returns: boolean
        '''
        filename = os.path.basename(file_url)
        if not os.path.exists(TMP_LOAD_FILE):
            os.mkdir(TMP_LOAD_FILE)
        localpath = os.path.join(TMP_LOAD_FILE, filename)
        flag = False
        
        try:
            r = urlopen(file_url)
            with open(localpath, "wb") as wd:
                while True:
                    s = r.read(1024*1024*10)
                    if not s:
                        break
                    wd.write(s)
            flag = self.push_file(bundle_id, localpath, remotepath)
        except Exception:
            logger.get_logger("xctest_%s" % self.udid).exception("download_file_and_push")
        finally:
            if os.path.exists(localpath):
                os.remove(localpath)
            return flag
    
    @rpc_method
    def get_sandbox_file_content(self, bundle_id, file_path):
        '''获取sandbox中文本文件的内容
        
        :param bundle_id: 应用的bundle_id
        :type bundle_id: str
        :param file_path: 沙盒目录
        :type file_path: str
        '''
        return DT().get_sandbox_file_content(self.udid, bundle_id, file_path)

    @rpc_method
    def pull_file(self, bundle_id, remotepath, localpath='/tmp', is_dir=False, is_delete = False):
        '''拷贝手机中sandbox指定目录的文件到Mac本地
        
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param remotepath: sandbox上的目录或者文件，例如：/Library/Caches/test/
        :type remotepath: str        
        :param localpath: 本地的目录
        :type localpath: str
        :param device_udid: 设备的udid
        :type device_udid: str
        :param is_dir: remotepath是否为目录，默认为单个文件
        :type is_dir: bool
        :param is_delete: 是否删除
        :type is_delete: bool
        :returns: list or None
        '''        
        return DT().pull_file(self.udid, bundle_id, remotepath, localpath, is_dir, is_delete)
    
    @rpc_method
    def list_files(self, bundle_id, file_path):
        '''列出手机上指定app中的文件或者目录
        
        :param bundle_id: app的bundle id
        :type bundle_id: str 
        :param file_path: sandbox的目录，例如: /Documents/
        :type file_path: str  
        '''
        response = self.agent.execute(Command.QTA_SANDBOX_LIST, {'path':file_path})
        if not isinstance(response, dict) or response['status'] == ErrorCode.QT4I_STUB_NOT_EXIST:
            file_list = DT().get_sandbox_path_files(self.udid, bundle_id, file_path)
            if not DT().is_simulator(self.udid):
                DT().close_sandbox_client()
        else:
            file_list = response['value']
        return file_list
    
    @rpc_method
    def remove_files(self, bundle_id, file_path):
        '''删除手机上指定app中的文件或者目录
        
        :param bundle_id: app的bundle id
        :type bundle_id: str 
        :param file_path: 待删除的文件或者目录，例如: /Documents/test.log
        :type file_path: str  
        '''
        response = self.agent.execute(Command.QTA_SANDBOX_REMOVE, {'path':file_path})
        if not isinstance(response, dict) or response['status'] == ErrorCode.QT4I_STUB_NOT_EXIST:
            DT().remove_files(self.udid, bundle_id, file_path)
    
    @rpc_method
    def get_syslog(self, watchtime, processName=None):
        '''获取手机系统日志
        
        :parm watchtime : 观察时间
        :type watchtime : int
        :parm processName : 手机服务名称(默认为全部查看)
        :type processName : str
        :returns: str
        '''
        if not os.path.isdir(TMP_DIR_PATH):
            os.mkdir(TMP_DIR_PATH)
        logFile = os.path.join(TMP_DIR_PATH, "sys_%s.log" %self.udid)         
        DT().get_syslog(watchtime, logFile, processName, self.udid)
        log = ''
        with open(logFile, 'rb') as fd:
            for line in fd:
                log += line
        return log

    @rpc_method
    def get_foreground_app_name(self):
        return self.agent.execute(Command.QTA_GET_FOREGROUND_APP_NAME)['value']
    
    @rpc_method
    def get_foreground_app_pid(self):
        return self.agent.execute(Command.QTA_GET_FOREGROUND_APP_PID)['value']
    
    @rpc_method
    def get_screen_orientation(self):
        orientation = self.agent.execute(Command.GET_SCREEN_ORIENTATION)['value']
        orientation_number = 0
        if orientation == 'PORTRAIT':
            orientation_number = 1
        elif orientation == 'UIA_DEVICE_ORIENTATION_PORTRAIT_UPSIDEDOWN':  
            orientation_number = 2
        elif orientation == 'LANDSCAPE':
            orientation_number = 3    
        elif orientation == 'UIA_DEVICE_ORIENTATION_LANDSCAPERIGHT': 
            orientation_number = 4
        return  orientation_number 
    
    @rpc_method
    def get_app_list(self, app_type):
        return DT().list_apps(self.udid, app_type)
    
    @rpc_method
    def get_sandbox_path_files(self, bundle_id, file_path):
        '''返回真机或者模拟器的沙盒路径
        
        :param bundle_id: 应用的bundle_id
        :type bundle_id: str
        :param file_path: 沙盒目录
        :type file_path: str
        '''
        return DT().get_sandbox_path_files(self.udid, bundle_id, file_path)

    @rpc_method
    def is_sandbox_path_dir(self, bundle_id, file_path):
        '''判断一个sandbox路径是否是一个目录
        
        :param bundle_id: 应用的bundle_id
        :type bundle_id: str
        :param file_path: 沙盒目录
        :type file_path: str
        '''
        return DT().is_sandbox_path_dir(self.udid, bundle_id, file_path)
    
    @rpc_method
    def close_sandbox_client(self):
        '''销毁sandboxClient对象
        '''
        DT().close_sandbox_client()
        
    @rpc_method
    def upload_photo(self, photo_data, album_name, cleared=True):
        '''上传照片到系统相册
        
        :param photo_data: base64编码后的数据
        :type photo_data: str
        :param album_name: 系统相册名
        :type album_name : str
        :param cleared: 是否清空已有的同名相册
        :type cleared : boolean
        :returns: boolean
        '''
        params = {'image':photo_data, 'albumName':album_name, 'cleared':cleared}
        response = self.agent.execute(Command.QTA_ALBUM_UPLOAD, params)
        if not isinstance(response, dict) or response['status'] == ErrorCode.QT4I_STUB_NOT_EXIST:
            return False
        else:
            return response['value']
        
    @rpc_method
    def call_qt4i_stub(self, method, params, clazz=None):
        '''QT4i Stub通用接口
        
        :param method: 函数名
        :type method : str
        :param params: 函数参数
        :type params : list
        :returns: str
        '''
        if clazz:
            params = {'method':method, 'params':params, 'object':clazz}
        else:
            params = {'method':method, 'params':params}
        
        if method.startswith('createFile'):
            url = "%s/stub" % self.agent.stub_server_url
            request = Request(url, data=json.dumps(params))
            response = urlopen(request)
            result = response.read() 
            result = json.loads(result)
            return result['result']
        
        response = self.agent.execute(Command.QTA_STUB_CALL, params)
        if not isinstance(response, dict) or response['status'] == ErrorCode.QT4I_STUB_NOT_EXIST:
            raise Exception('QT4i Stub not exist')
        else:
            res = response['value']
            if res['result']:
                return res['data']
            else:
                raise Exception('method not implemented in QT4i Stub')
            
    @rpc_method
    def get_device_detail(self):
        '''获取设备详细信息
        '''
        is_simulator = DT().is_simulator(self.udid)
        device_type = self.agent.execute(Command.QTA_DEVICE_DETAIL_INFO)['value']
        return DeviceProperty.get_property(device_type, is_simulator)


class Application(RPCEndpoint):
    '''UIAApplication API
    '''
    rpc_name_prefix = "app."

    def __init__(self, rpc_server, device_id):
        self.rpc_server = rpc_server
        self.udid = device_id
        RPCEndpoint.__init__(self)
    
    @lazy
    def agent(self):
        '''延迟初始化agent
        '''
        agent_manager = XCUITestAgentManager()
        if agent_manager.has_agent(self.udid):
            return agent_manager.get_agent(self.udid)
        else:
            return None

    @rpc_method
    def get_app_bundle_id(self):
        '''获取当前App的BundleID
        
        :returns: str
        '''
        if self.agent and self.agent.has_session():
            return self.agent.capabilities['CFBundleIdentifier']
    
    @rpc_method
    def get_app_version(self):
        '''获取App的Version
        
        :returns: str
        '''
        raise NotImplementedError

    @rpc_method
    def has_crashed(self): 
        '''当前App是否发生crash
        
        :returns: boolean
        '''
        if self.agent:
            return self.agent.get_crash_flag()
        return False
    

class Element(RPCEndpoint):
    '''UIAElement API
    '''
    rpc_name_prefix = "element."

    def __init__(self, rpc_server, device_id):
        self.rpc_server = rpc_server
        self.udid = device_id
        RPCEndpoint.__init__(self)
    
    @lazy
    def agent(self):
        '''延迟初始化agent
        '''
        agent_manager = XCUITestAgentManager()
        if agent_manager.has_agent(self.udid):
            return agent_manager.get_agent(self.udid)
        else:
            return None
    
    @rpc_method
    def find_element(self, locator, timeout=3, interval=0.005, strategy=By.QPATH, parent_id=None):
        '''查找单个element对象，返回查找到的element对象的缓存id，不指定parent_id则从Application下开始搜索。
        
        :param locator: 定位element的字符串，例如: "/classname='UIAWindow'"
        :type locator: str
        :param timeout: 查找element的超时值（单位: 秒）
        :type timeout: float
        :param interval: 在查找element的超时范围内，循环的间隔（单位: 秒）
        :type interval: float
        :param strategy: locator的类型，例如: "qpath", "xpath", "id", "name"
        :type strategy: str
        :param parent_id: 父element的id，如不指定则从UIAApplication下查找
        :type parent_id: str
        :returns: dict : {'element': id, 'path': <encode str>, 'valid_path_part': <encode str>, 'invalid_path_part': <encode str>, 'find_count': <int>, "find_time": int}
        '''
        result = self.find_elements(locator, timeout, interval, strategy, parent_id)
        if result['elements']:
            result.update(result['elements'][0])
        else:
            result['element'] = None
        result.pop('elements')
        return result
    
    @rpc_method
    def find_elements(self, locator, timeout=3, interval=0.005, strategy=By.QPATH, parent_id=None):
        '''查找单个element对象，返回查找到的element对象的缓存id。不指定parent_id则从UIAApplication下开始搜索。
        
        :param locator: 定位element的字符串，例如: "/classname='UIAWindow'"
        :type locator: str
        :param timeout: 查找element的超时值（单位: 秒）
        :type timeout: float
        :param interval: 在查找element的超时范围内，循环的间隔（单位: 秒）
        :type interval: float
        :param strategy: locator的类型，例如: "qpath", "xpath", "id", "name"
        :type strategy: str
        :param parent_id: 父element的id，如不指定则从UIAApplication下查找
        :type parent_id: int
        :returns: dict : {'elements': [{'element':id, 'attributes':<encode str>}], 'path': <encode str>, 'valid_path_part': <encode str>, 'invalid_path_part': <encode str>, 'find_count': <int>, "find_time": int}
        '''
        if strategy == "qpath":
            #解析QPath
            qpath_array, qpath_lex = QPathParser().parse(locator)
            for node in qpath_array:
                if 'classname' in node:
                    if node['classname'][1] in UIA_XCT_MAPS.keys():
                        node['classname'][1] = UIA_XCT_MAPS[node['classname'][1]]
            locator_value = qpath_array
            valid_path = None
            invalid_path = None
        else:
            locator_value = locator
        start_time = time.time()
        elements = []
        count = 0
        parent_id = parent_id if parent_id else 1
        while True:
            before_step_time = time.time()
            count += 1
            response = self.agent.execute(Command.QTA_FIND_ELEMENTS,
                                {'using': strategy, 'value': locator_value, 'id':parent_id})
            elements = response['value']
            after_step_time = time.time()
            step_time = after_step_time - before_step_time
            if elements or after_step_time - start_time >= timeout:
                break
            if interval > 0:
                sleep_time = interval - step_time
                if sleep_time > 0:
                    time.sleep(sleep_time)
        if elements:
            elements = [{'element': int(e['ELEMENT'])} for e in elements]
            if strategy == "qpath":
                valid_path = locator
                invalid_path = ""
        else:
            if strategy == "qpath":
                valid_path_index = response['info']['index']
                locations = []
                if valid_path_index + 1 < len(qpath_lex):
                    valid_path_index = valid_path_index + 1
                for _ in qpath_lex[valid_path_index].values():
                    locations.extend(_)
                locations.sort()
                valid_index = locations[0] - 1 
                valid_path = locator[0:valid_index]
                invalid_path = locator[valid_index+1:]
        find_time = int((time.time() - start_time)*1000)
        result = {'elements':elements, 
                  'find_count':count,
                  'find_time':find_time}
        if strategy == "qpath":
            result['valid_path_part'] = valid_path
            result['invalid_path_part'] = invalid_path
        return result

    @rpc_method
    def find_element_with_predicate(self, element_id, predicate):
        '''通过predicate文本获取第一个匹配的子element
        
        :param element_id: element的id
        :type element_id: int
        :param predicate: 预期子element的predicate （例如：“name beginswith 'xxx'”）
        :type predicate: str
        :returns: int or None
        '''
        try:
            element = self.agent.execute(Command.FIND_CHILD_ELEMENT, {'id':element_id, 'using':By.PREDICATE_STRING, 'value':predicate})['value']
            if 'ELEMENT' in element:
                return int(element['ELEMENT'])
        except NoSuchElementException:
            return

    @rpc_method
    def find_elements_with_predicate(self, element_id, predicate):
        '''通过predicate文本获取所有匹配的子element
        
        :param element_id: element的id
        :type element_id: int
        :param predicate: 预期子element的predicate （例如：“name beginswith 'xxx'”）
        :type predicate: str
        :returns: list - [id1, id2, id3...]
        '''
        try:
            elements = self.agent.execute(Command.FIND_CHILD_ELEMENTS, {'id':element_id, 'using':By.PREDICATE_STRING, 'value':predicate})['value']
            if type(elements) == type([]):
                return [{'element': int(e['ELEMENT'])} for e in elements]
        except NoSuchElementException:
            return []

    @rpc_method
    def find_element_with_value_for_key(self, element_id, key, value):
        '''通过匹配指定key的value，获取第一个匹配的子element
        
        :param element_id: element的id
        :type element_id: int
        :param key: key （例如：label、name、value）
        :type key: str
        :param value: 对应key的value值
        :type value: str
        :returns: int or None
        '''
        try:
            element = self.agent.execute(Command.FIND_CHILD_ELEMENT, {'id':element_id, 'using':By.LINK_TEXT, 'value':'%s=%s' %(key, value)})['value']
            if 'ELEMENT' in element:
                return int(element['ELEMENT'])
        except NoSuchElementException:
            return

    @rpc_method
    def find_elements_with_value_for_key(self, element_id, key, value):
        '''通过匹配指定key的value，获取所有匹配的子element
        
        :param element_id: element的id
        :type element_id: int
        :param key: key （例如：label、name、value）
        :type key: str
        :param value: 对应key的value值
        :type value: str
        :returns: list - [id1, id2, id3...]
        '''
        try:
            elements = self.agent.execute(Command.FIND_CHILD_ELEMENTS, {'id':element_id, 'using':By.LINK_TEXT, 'value':'%s=%s' %(key, value)})['value']
            if type(elements) == type([]):
                return [{'element': int(e['ELEMENT'])} for e in elements]
        except NoSuchElementException:
            return []
    
    @rpc_method
    def get_element_tree(self, element_id, max_depth=0):
        '''获取指定element的所有子孙element树
        
        :param element_id: element的id
        :type element_id: int
        :param max_depth: 树的深度, 默认为0，获取所有子孙控件树
        :type max_depth: int
        :returns: dict
        '''
        if element_id == 1:
            result = self.agent.execute(Command.GET_ELEMENT_TREE)['value']['tree']
        else:
            result = self.agent.execute(Command.QTA_ELEMENT_TREE, 
                {'id': element_id, 'max_depth':max_depth})['value']['tree']
        return convert_to_qpath(result)
    
    @rpc_method
    def get_parent_element(self, element_id):
        '''获取指定element的父element的id
        
        :param element_id: element的id
        :type element_id: int
        :returns: int
        '''
        if element_id == 1:
            raise Exception("Current element is already root element.")
        element =  self.agent.execute(Command.QTA_GET_PARENT_ELEMENT, {'id': element_id})['value']
        if 'ELEMENT' in element:
            return int(element['ELEMENT'])
    
    @rpc_method
    def get_children_elements(self, element_id):
        '''获取指定element的子elements集合
        
        :param element_id: element的id
        :type element_id: int
        :returns: list : [int, ...]
        '''
        elements = self.agent.execute(Command.QTA_GET_CHILDREN_ELEMENTS, {'id': element_id})['value']
        if type(elements) == type([]):
            return [int(e['ELEMENT']) for e in elements]
        return []
    
    @rpc_method
    def get_rect(self, element_id):
        '''获取指定element的坐标和长宽
        
        :param element_id: element的id
        :type element_id: int
        :return: dict : 坐标和长宽
        '''
        if element_id == 1:
            format_rect = {
                'origin': {'x': 0, 'y': 0},
                'size'  : self.agent.execute(Command.GET_WINDOW_SIZE)['value'],
            }
        else:
            rect = self.agent.execute(Command.GET_ELEMENT_RECT, {'id': element_id})['value']
            format_rect = {
                'origin': {'x': rect['x'], 'y': rect['y']},
                'size'  : {'width': rect['width'], 'height': rect['height']},
            }
        return format_rect
    
    @rpc_method
    def capture(self, element_id, filepath=None):
        '''截图（截取指定element的图片，并将图片输出至指定的路径）
        
        :param element_id: element的id
        :type element_id: int
        :param filepath: 将图片存储至该路径（png格式），不传路径 或 路径仅含文件名，则截图后存储至默认路径。
        :type filepath: str
        :returns: str
        '''
        raise NotImplementedError
    
    @rpc_method
    def click(self, element_id, x=0.5, y=0.5, options=None):
        '''点击指定的element
        
        :param element_id: element的id
        :type element_id: int
        :param x: x坐标（相对于当前element），可不传入该参数
        :type x: float
        :param y: y坐标（相对于当前element），可不传入该参数
        :type y: float
        '''
        if x is None:
            x = 0.5
        if y is None:
            y = 0.5
        if element_id == 1:
            self.agent.execute(Command.QTA_DEVICE_CLICK, {'x':x, 'y':y, 'options':options})['value']
        else:     
            self.agent.execute(Command.QTA_ELEMENT_CLICK, 
                {'id':element_id, 'x':x, 'y':y, 'options':options})['value']
    
    @rpc_method
    def double_click(self, element_id, offset_x=0.5, offset_y=0.5):
        '''双击指定的element
        
        :param element_id: element的id
        :type element_id: int
        '''
        if element_id == 1:
            self.agent.execute(Command.QTA_DEVICE_DOUBLE_CLICK, {'x':0.5, 'y':0.5})['value']
        else:
            self.agent.execute(Command.QTA_ELEMENT_DOUBLE_CLICK, {'id':element_id, 'x':offset_x, 'y':offset_y})['value']
    
    @rpc_method
    def drag(self, element_id, x0, y0, x1, y1, duration=0.1):
        '''控件内部拖拽
        
        :param element_id: element的id
        :type element_id: int
        :param x0: 起始横向坐标（从左向右，屏幕百分比）
        :type x0: float
        :param y0: 起始纵向坐标（从上向下，屏幕百分比）
        :type y0: float
        :param x1: 终止横向坐标（从左向右，屏幕百分比）
        :type x1: float
        :param y1: 终止纵向坐标（从上向下，屏幕百分比）
        :type y1: float
        :param duration: 起始坐标按下的时间（秒）
        :type duration: float
        '''
        if element_id == 1:
            self.agent.execute(Command.QTA_DEVICE_DRAG, 
                {'x0':x0, 'y0':y0, 'x1':x1, 'y1':y1,'duration':duration})['value']
        else:
            self.agent.execute(Command.QTA_ELEMENT_DRAG, 
                {'id':element_id, 'x0':x0, 'y0':y0, 'x1':x1, 'y1':y1,'duration':duration})['value']
    
    @rpc_method
    def tap(self, element_id, options):
        '''拂去/拖拽等自定义操作
        
        :param element_id: element的id
        :type element_id: int
        :param options     : 自定义项
        :type options      : dict : {
                            :           touchCount   : 1,                    // 触摸点
                            :           duration     : 0.5,                  // 时间
                            :           startOffset  : { x: 0.0, y: 0.1},    // 偏移百分比
                            :           endOffset    : { x: 1.0, y: 0.1 },   // 偏移百分比
                            :           repeat       : 1                     // 重复该操作
                            :           interval     : 0                     // 重复该操作的间隙时间（秒）
                            :        }
        '''
        raise NotImplementedError
    
    @rpc_method
    def rotate(self, element_id, options):
        '''旋转
        
        :param element_id: element的id
        :type element_id: int
        :param _options     : 自定义项
        :type _options      : dict : {
                            :           centerOffset : {x:0.0, y:0.0},   // 中心点
                            :           duration     : 1.5,              // 持续时间
                            :           radius       : 100,              // 半径
                            :           rotation     : 100,              // 旋转弧度的长度，默认为圆周率PI
                            :           touchCount   : 2                 // 触摸点（最大5个点，这里默认2个）
                            :          }
        '''
        raise NotImplementedError
    
    @rpc_method
    def scroll_to_visible(self, element_id, rate=1.0):
        '''自动滚动到该element为可见
        
        :param element_id: element的id
        :type element_id: int
        '''
        self.agent.execute(Command.QTA_SCROLL_TO_VISIBLE, {'id':element_id, 'velocity':rate})
    
    @rpc_method
    def long_click(self, element_id, offset_x=0.5, offset_y=0.5, seconds=3):
        '''持续按住
        
        :param element_id: element的id
        :type element_id: int
        :param seconds: 秒
        :type seconds: int
        '''
        if element_id == 1:
            self.agent.execute(Command.QTA_DEVICE_LONG_CLICK, {'x':0.5, 'y':0.5, 'duration':seconds})['value']
        else:
            self.agent.execute(Command.QTA_ELEMENT_LONG_CLICK, {'id':element_id, 'x':offset_x, 'y':offset_y, 'duration':seconds})['value']
    
    @rpc_method
    def wait_for_invalid(self, element_id, seconds=5):
        '''等待当前element直到无效
        
        :param element_id: element的id
        :type element_id: int
        :param seconds: 秒
        :type seconds: int
        :returns: boolean
        '''
        raise NotImplementedError
    
    @rpc_method
    def set_value(self, element_id, value):
        '''设置指定element的value
        
        :param element_id: element的id
        :type element_id: int
        :param value: 值
        :type value: str
        '''
        self.agent.execute(Command.QTA_ELEMENT_SETVALUE, {'id':element_id, 'value':value})['value']
    
    @rpc_method
    def drag_to_value(self, element_id, value):
        '''slider滑到指定位置(0-1之间)
        
        :param element_id: element的id
        :type element_id: int
        :param value: 值
        :type value: str
        '''
        self.agent.execute(Command.QTA_DRAG_TO_VALUE, {'id':element_id, 'value':value})
        
    @rpc_method
    def send_keys(self, element_id, keys):
        '''在指定element中键盘输入
        
        :param element_id: element的id
        :type element_id: int
        :param keys: 键入的字符串
        :type keys: str
        '''
        if element_id == 1 :
            return self.agent.execute(Command.QTA_DEVICE_SENDKEYS, {'value':keys})['value']
        else:
            return self.agent.execute(Command.QTA_ELEMENT_SENDKEYS, {'id':element_id, 'value':keys})['value']
    
    @rpc_method
    def get_element_attrs(self, element_id):
        '''获取指定element的属性信息
        
        :param element_id: element的id
        :type element_id: int
        :returns: str: dict
        '''
        attrs = self.agent.execute(Command.QTA_GET_ELEMENT_ATTRS, {'id': element_id})['value']
        return self._convert_attrs(attrs)
    
    def _convert_attrs(self, attrs):
        '''转译属性名称及对应值
        '''
        if 'type' in attrs:
            attrs['classname'] = attrs['type']
            attrs.pop('type')
        if 'isEnabled' in attrs:
            attrs['enabled'] = (attrs['isEnabled'] == '1')
            attrs.pop('isEnabled')
        if 'isVisible' in attrs:
            attrs['visible'] = (attrs['isVisible'] == '1')
            attrs.pop('isVisible')
        children = []
        for child in attrs.get('children', []):
            children.append(self._convert_attrs(child))
        attrs['children'] = children
        return attrs
    
    @rpc_method
    def get_element_attr(self, element_id, attr_name):
        '''获取指定element的属性信息
        
        :param element_id: element的id
        :type element_id: int
        :param attr_name: 属性名
        :type attr_name: str
        :returns: str or boolean or None
        '''
        return self.agent.execute(Command.GET_ELEMENT_ATTRIBUTE, {'id': element_id, 'name': attr_name})['value']
    
    @rpc_method
    def get_element_win_info(self, element_id):
        '''获取指定element的窗口信息
        
        :param element_id: element的id
        :type element_id: int
        :returns: dict
        '''
        return self.agent.execute(Command.QTA_GET_ELEMENT_WIN_INFO, {'id': element_id})['value']
    
    @rpc_method
    def select_picker_wheel(self, element_id, value):
        '''设置PickerWheel的选项
        
        :param element_id: element的id
        :type element_id: int
        :param value: 选项
        :type value: str
        '''
        self.agent.execute(Command.QTA_WHEEL_SELECT, {'id':element_id, 'value':value})['value']

    @rpc_method
    def force_touch(self, element_id, pressure, duration):
        self.agent.execute(Command.FORCE_TOUCH, 
            {'id':element_id, 'pressure':pressure, 'duration':duration})['value']


# coding: UTF-8
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''ios_driver
'''
# -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*-
from __future__ import absolute_import, print_function

import os, re, time, urllib
from qt4i.driver.instruments.internal._xcode   import Xcode
from qt4i.driver.instruments.internal._timer   import ThreadTimer
from qt4i.driver.util import Args
from qt4i.driver.util import FileManager
from qt4i.driver.instruments.internal._jsonp import Json
from qt4i.driver.util import Process
from qt4i.driver.util import Task
from qt4i.driver.tools import logger

from qt4i.driver.rpc import RPCClientProxy
# -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*-

class Instruments(object):
    
    CurrentDirectoryPath = os.path.abspath(os.path.dirname(__file__))
    PythonPath = Task('which python').execute()
    
    def _clean_instrumentscli_n_trace(self, trace_output=None, dir_path=CurrentDirectoryPath):
        if trace_output:
            FileManager(trace_output).force_delete(trace_output)
        items = FileManager(path=dir_path).get_children()
        for item in items:
            if re.match('^instrumentscli\d+.trace$', os.path.basename(item)):
                FileManager(path=item).force_delete()
    
    def _clean_uia_results_path(self):
        manager = FileManager(self.environment.get('uia_results_path'))
        if manager.exists():manager.force_delete()
    
    def _renew_output_path(self):
        for item in ['uia_results_path', 'screen_shot_path']:
            manager = FileManager(self.environment.get(item))
            if manager.exists():manager.force_delete()
            manager.repair_path()
    
    def _update_environment_file(self):
        with open(os.path.join(self.CurrentDirectoryPath, '_environment.js'), 'wb') as fd:
            fd.write('var Environment = %s;' % Json().dumps(self.environment, indent=4))
    
    def _find_instruments_uia_result_path(self):
        if os.path.exists(self.environment.get('uia_results_path')):
            items = FileManager(path=self.environment.get('uia_results_path')).get_children(sort_by_ctime=True)
            if len(items) > 0 : return items[-1]
            else:self.logger.error('Warning: instruments_uia_result_path not found.')
    
    def _copy_instrumentscli_n_trace(self, dir_path=CurrentDirectoryPath):
        dst = self.environment.get('trace_output')
        if dst:
            items = FileManager(path=dir_path).get_children()
            for item in items:
                if re.match('^instrumentscli\d+.trace$', os.path.basename(item)):
                    fm = FileManager(path=item)
                    fm.force_copy_to(dst)
                    fm.force_delete()
    
    def __init__(self, **args):
#         self.logger = Logger(os.path.join('/tmp', 'instruments.log'))
        self.logger = logger.get_logger("instruments")
        self.args = args
        if self.args.get('device_udid') is None: raise Exception('device_udid is None.')
        if self.args.get('device_udid') is None: raise Exception('bundle_id is None.')
        self.device_udid = self.args['device_udid']
        self.environment = {'device_udid'                 : self.args['device_udid'],
                            'device_simulator'            : self.args.get('device_simulator', False),
                            'bundle_id'                   : self.args['bundle_id'],
                            'trace_template'              : self.args.get('trace_template'),
                            'trace_output'                : self.args.get('trace_output'),
                            'uia_script'                  : os.path.abspath(self.args.get('uia_script', os.path.join(self.CurrentDirectoryPath, '_bootstrap.js'))),
                            'uia_results_path'            : os.path.join('/tmp', '_uiaresults'),
                            'screen_shot_path'            : os.path.join('/tmp', '_screenshot'),
                            'cwd'                         : self.CurrentDirectoryPath,
                            'timeout'                     : self.args.get('timeout', 20 * 60),
                            'xmlrpc_uri'                  : self.args.get('xmlrpc_uri', '/'.join(['http://0.0.0.0:12306', 'device', '%s'%self.device_udid])),
                            'cmd_fetch_delegate'          : os.path.join(self.CurrentDirectoryPath, '_cmd_fetch_delegate.py'),
                            'cmd_fetch_delegate_timeout'  : self.args.get('cmd_fetch_delegate_timeout', 10),
                            'ignore_cmd_error'            : self.args.get('ignore_cmd_error', True),
                            'python_path'                 : self.PythonPath}
        self.environment.update(eval(urllib.unquote(self.args.get('environment'))))
        self.args.update(self.environment)
        self.args.update({'started_callback': self.__started_callback__, 'stdout_line_callback': self.__stdout_line_callback__, 'stderr_line_callback': self.__stderr_line_callback__, 'return_code_callback': self.__return_code_callback__})
        self._clean_instrumentscli_n_trace(self.environment.get('trace_output'))
        self._renew_output_path()
        self._update_environment_file()
        self.out = list()
        self.err = list()
        self.rpc = RPCClientProxy(self.environment.get('xmlrpc_uri'), allow_none=True)
        self.instruments = Xcode().start_instruments(**self.args)
        self.logger.info(self.instruments.command)
        self.instruments_uia_result_path = None
        self.logger.info("Instruments timeout %d"%self.environment.get('timeout'))
        self.instruments_timeout_daemon = ThreadTimer(timeout=self.environment.get('timeout'), callback=self.__timeout_kill_instruments__)
        self.instruments_pid = self.instruments.pid
        self.instruments_trace_complete = False
        self.instruments.wait()
    
    def _get_out(self):
        if len(self.out) > 0:
            return ''.join(self.out)
    
    def _get_err(self):
        if len(self.err) > 0:
            return ''.join(self.err)
    
    def _repair_datetime_prefix(self, line):
        datetime_string, line = None, line
        datetime_pattern = '^\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}\:\d{2}'
        datetime_format = '%Y-%m-%d %H:%M:%S'
        datetime_timezone = 60 * 60 * 8
        if re.search('%s \+0{4} ' % datetime_pattern, line):
            datetime_string, line = re.split('(?<=%s) \+0{4} \w+\: |(?<=%s) \+0{4} ' % (datetime_pattern, datetime_pattern), line, maxsplit=1)
            datetime_string = time.strftime(datetime_format, time.localtime(time.mktime(time.strptime(datetime_string, datetime_format)) + datetime_timezone))
        return datetime_string, line
    
    def __notify__(self, method):
        self.logger.info(method)
        try: self.rpc.ins.notify(Json().dumps({"method": "%s" % method, "params": [{'pid': self.instruments_pid, 'environment': self.environment, 'error': self._get_err()}]}))
        except Exception as e: self.logger.error(str(e))
    
    def __timeout_kill_instruments__(self):
        self.instruments.stop()
        Process().kill_child_processes_by_ppid(os.getpid())
        self.__notify__('InstrumentsTimeout')
    
    def __started_callback__(self, pid):
        self.instruments_pid = pid
        self.__notify__('InstrumentsStarted')
    
    def __stdout_line_callback__(self, line):
        datetime, line = self._repair_datetime_prefix(line)
        if datetime : self.logger.info('[%s][%s] %s' % (datetime, os.getpid(), line)),
        else        : self.logger.info(line),
        self.out.append(line)
        self.__excavate__(line)
        if self.instruments_timeout_daemon : self.instruments_timeout_daemon.cancel()
        self.instruments_timeout_daemon = ThreadTimer(timeout=self.environment.get('timeout'), callback=self.__timeout_kill_instruments__)
    
    def __stderr_line_callback__(self, line):
        datetime, line = self._repair_datetime_prefix(line)
        if datetime : self.logger.info('[%s][%s] %s' % (datetime, os.getpid(), line)),
        else        : self.logger.info(line),
        self.err.append(line)
        self.__excavate__(line)
    
    def __return_code_callback__(self, pid, returncode):
        self.instruments_pid = pid
#         self._copy_instrumentscli_n_trace()  #直接在指定目录生成trace，无需拷贝
        self._clean_instrumentscli_n_trace()
        self._clean_uia_results_path()
        self.__notify__('InstrumentsTerminated')
    
    def __excavate__(self, line):
        if self.instruments_trace_complete == False and re.match('^ScreenshotCaptured:.+$', line):
            if self.instruments_uia_result_path is None:
                self.instruments_uia_result_path = self._find_instruments_uia_result_path()
            if self.instruments_uia_result_path is not None:
                data = Json().loads(re.sub('^ScreenshotCaptured:', '', line))
                src = os.path.join(self.instruments_uia_result_path, data['name'])
                dst = data['path']
                start_time = time.time()
                while time.time() - start_time < 10:
                    if os.path.exists(src):
                        FileManager(src).force_copy_to(dst)
                        if os.path.exists(dst): return
                    time.sleep(0.03)
                self.logger.error('Warning: screenshot failed [ %s ]' % src)
                return
        if self.instruments_trace_complete == False and re.match('^TORPC:.+$', line):
            try: self.rpc.ins.send_result(re.sub('^TORPC:', '', line))
            except Exception as e: self.logger.error(str(e))
            return
        if self.instruments_trace_complete == False and re.match('^Instruments Usage Error : The specified template .+$', line, re.I):
            self.__notify__('TraceTemplateInvalid')
            self.instruments_trace_complete = True
            return
        if self.instruments_trace_complete == False and re.match('^Instruments Usage Error : Specified target process is invalid: .+$', line, re.I):
            self.__notify__('TargetAppBundleIDInvalid')
            self.instruments_trace_complete = True
            return
        if self.instruments_trace_complete == False and re.match('^Instruments Trace Error : Target failed to run: Device is currently locked with a passcode.  Please unlock the device and try again.$', line, re.I):
            self.__notify__('DeviceIsCurrentlyLockedWithAPasscode')
            self.instruments_trace_complete = True
            return
        if self.instruments_trace_complete == False and re.match('^Instruments Trace Error : Target failed to run: Permission to debug com.tencent.qq.dailybuild.test.gn was denied.  The app must be signed with a development identity \(e.g. iOS Developer\).$', line, re.I):
            self.__notify__('TheAppMustBeSignedWithADevelopmentIdentity')
            self.instruments_trace_complete = True
            return
        if self.instruments_trace_complete == False and re.match('^Could not start script, target application is not frontmost.$', line, re.I):
            self.__notify__('TargetAppIsNotFrontmost')
            self.instruments_trace_complete = True
            return
        if self.instruments_trace_complete == False and re.match('^The target application appears to have died$', line, re.I):
            self.__notify__('TargetAppDied')
            self.instruments_trace_complete = True
            return
        if self.instruments_trace_complete == False and re.match('^.+WebKit Threading Violation - initial use of WebKit from a secondary thread.$', line, re.I):
            self.__notify__('DeviceDisabled')
            return
        if self.instruments_trace_complete == False and re.match('^Instruments Usage Error :.+$', line, re.I):
            self.__notify__('InstrumentsError')
            self.instruments_trace_complete = True
            return
        if self.instruments_trace_complete == False and re.match('^Instruments Trace Complete.+$', line, re.I):
            self.instruments.keep_running = False
            self.instruments_trace_complete = True
            return

# -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*-

if __name__ == '__main__':

    # ***** ***** ***** ***** ***** ***** ***** *****
    # 参数命令:（参数名不区分大小写，参数值区分大小写）
    #   device_udid                 - 设备UUID
    #   device_simulator            - 设备是否模拟器
    #   bundle_id                   - APP的BundleID
    #   uia_script                  - 脚本
    #   # 以下参数无特殊情况可不传入
    #   trace_template              - 定制化的trace_template
    #   trace_output                - 定制化的trace_output
    #   timeout                     - 闲置超时（单位:秒。默认为1200秒=20分钟），如果持续长时间无运行状态，达到执行的超时值则终止持续运行
    #   xmlrpc_uri                  - XmlRpcServer的uri（默认值: http://0.0.0.0:11516）
    #   cmd_fetch_delegate_timeout  - 请求命令的超时值（单位：秒。只有特殊情况才修改该值）
    #   ignore_cmd_error            - cmd执行异常后，输出异常但不终止程序，默认为True（由应用方判定是否终止程序）
    # ***** ***** ***** ***** ***** ***** ***** *****

    Instruments(**Args())



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''Process扩展
'''


import os
import re
import signal
import subprocess
import six


class Process(object):

    def get_processes_by_port(self, port, state="LISTEN"):
        items = []
        
        if isinstance(port, int):
            cmd = r"lsof -nP -iTCP:%s  -sTCP:%s | awk -s '{print  $2}'" % (port, state)
            items = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            if six.PY3:
                items = items.decode()
            items = items.split('\n')
            items = [item for item in items if re.match(r"^\d+$", item)]
        return items

    def get_processes_by_grep(self, grep=''):
        items = []
        stdout = os.popen('ps -ef').read()
        if len(stdout) > 0:
            lines = stdout.split('\n')
            top_line = re.sub('^ +| $', '', lines[0])
            top_line = re.sub(' +', ' ', top_line)
            keys = top_line.split(' ')
            if len(keys) != 8: raise Exception('KeysError')
            lines = lines[1:]
            if isinstance(grep, six.string_types): lines = os.popen('ps -ef|grep "%s" | grep -v grep' % grep).read().split('\n')
            for line in lines:
                if len(line) > 0:
                    line = re.sub('^ +| $', '', line)
                    line = re.sub(' +', ' ', line)
                    values = re.split(' ', line, len(keys) - 1)
                    if len(values) != 8: raise Exception('ValuesError')
                    _dict = {}
                    for i in range(8):
                        key = keys[i]
                        value = values[i]
                        if re.match('^\d+$', value):
                            value = int(value)
                        _dict.update({key: value})
                    items.append(_dict)
        return items

    def get_process_by_pid(self, pid, grep=''):
        if isinstance(pid, int):
            processes = self.get_processes_by_grep(grep)
            for process in processes:
                if process.get('PID') == pid: return process

    def get_child_processes_by_ppid(self, ppid, grep=None):
        children = []
        if isinstance(ppid, int):
            processes = self.get_processes_by_grep(grep)
            for process in processes:
                if process.get('PPID') == ppid: children.append(process)
        return children

    def kill_process_by_pid(self, pid):
        if isinstance(pid, int):
            try:
                os.kill(pid, signal.SIGKILL)
            except:
                pass
        return self.get_process_by_pid(pid) is None

    def kill_process_tree_by_pid(self, pid, grep=None):
        result = True
        if isinstance(pid, int):
            processes = self.get_child_processes_by_ppid(pid, grep)
            for process in processes:
                if self.kill_process_tree_by_pid(process.get('PID')) is False:
                    result = False
            if (self.kill_process_by_pid(pid) and len(self.get_child_processes_by_ppid(pid, grep))==0) is False:
                result = False
        return result

    def kill_child_processes_by_ppid(self, ppid, grep=None):
        if isinstance(ppid, int):
            processes = self.get_child_processes_by_ppid(ppid, grep)
            for process in processes:
                self.kill_process_tree_by_pid(process.get('PID'))
        return len(self.get_child_processes_by_ppid(ppid, grep)) == 0

    def kill_process_by_port(self, port):
        pids = self.get_processes_by_port(port)
        current_pid = os.getpid()
        for pid in pids:
            try:
                if pid != str(current_pid):
                    os.kill(int(pid), signal.SIGKILL)
            except:
                pass
                    
    def kill_process_by_name(self, script_name): 
        cmd = "ps ax | grep '%s' | grep -v grep | awk -s '{print $1}'" % script_name
        result = True
        try:
            data = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            if data:
                if six.PY3:
                    data = data.decode()
                pids = data.strip().split("\n")
                for pid in pids:
                    os.kill(int(pid), signal.SIGKILL)
        except:
            result = False
        return result


# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''DeviceTools
'''

from __future__ import absolute_import, print_function

import fcntl 
import json
import os
import pkg_resources
import re
import subprocess
import shutil
import time
import datetime
import base64
from six.moves.http_client import HTTPConnection
from six import PY3
from six import with_metaclass
import six

from qt4i.driver.tools import mobiledevice
from qt4i.driver.tools.mobiledevice import InstallationProxy
from qt4i.driver.tools.sched import PortManager
from qt4i.driver.util import zip_decompress
from qt4i.driver.tools.mobiledevice import SandboxClient
from testbase.util import Singleton
from testbase.util import Timeout


# 挂载根路径
MOUNT_ROOT_PATH = os.path.expanduser('~')
TIME_FORMAT = '%Y-%m-%d %H:%M:%S'

class Flock(object):
    
    def __init__(self, filepath='/cores/lock'):
        self._filepath = filepath
        self._fd = open(self._filepath, 'w')

    def __enter__(self):
        fcntl.flock(self._fd, fcntl.LOCK_EX)
        with open('/cores/pid','w+') as fd:
            fd.write("%s\n" % str(os.getpid()))
        return self

    def __exit__(self, *args):
        fcntl.flock(self._fd, fcntl.LOCK_UN)
        try:  
            self._fd.close()  
            os.unlink(self._filepath)
        except:  
            pass
        self._fd = None 
    
    
def func_retry_wrap(func):
    def _wrap_func(*args, **kwargs):
        '''
        '''
        for _ in range(3):
            try:
                return func(*args, **kwargs)
            except:
                time.sleep(10)
        else:
            raise
    return _wrap_func

class DT(with_metaclass(Singleton, object)):
    '''
    DeviceTools
    '''
    
    FBSIMCTL_DEFAULT_PATH = '/cores/fbsimctl/fbsimctl'
    
    def __init__(self):
        self.xcode_version = DT.get_xcode_version()
        self.udid = None
        self.bundle_id = None
        self.sc = None
        self._init_fbsimctl()
    
    def _init_fbsimctl(self):
        try:
            fbsimctl_zip = pkg_resources.resource_filename("qt4i", "driver/tools/fbsimctl/fbsimctl.zip") #@UndefinedVariable
            if not os.path.exists(fbsimctl_zip):
                raise Exception('fbsimctl not found')
            self.fbsimctl = DT.FBSIMCTL_DEFAULT_PATH
        except:
            try:
                os.environ['PATH'] += ':/usr/local/bin'
                result = subprocess.check_output("which fbsimctl", env=os.environ, shell=True, stderr=subprocess.STDOUT, )
                if PY3:
                    result = result.decode()
                self.fbsimctl = result.split('\n')[0]
            except subprocess.CalledProcessError: 
                raise Exception('fbsimctl not found, use brew install')
        if self.fbsimctl == DT.FBSIMCTL_DEFAULT_PATH and not os.path.exists(self.fbsimctl): # 解压fbsimctl到默认路径
            fbsimctl_root_path = os.path.split(DT.FBSIMCTL_DEFAULT_PATH)[0]
            zip_decompress(fbsimctl_zip, fbsimctl_root_path)
            os.chmod(self.fbsimctl, 0o755)
            fbsimctl_fmk_path = os.path.split(DT.FBSIMCTL_DEFAULT_PATH)[0]
            for root, dirs, _files in os.walk(fbsimctl_fmk_path):
                if root.endswith('.framework'):
                    if 'Headers' not in dirs: # 为Framework添加Headers/Modules/Resources的外链
                        framework_name = os.path.splitext(os.path.basename(root))[0]
                        for item in ['Headers', 'Modules', 'Resources', framework_name]:
                            item_path = os.path.join(root, item)
                            if os.path.exists(item_path) and not os.path.islink(item_path):
                                os.remove(item_path)
                            os.symlink('Versions/Current/%s' % item, item_path)
                if root.endswith('Versions'):
                    if 'Current' not in dirs:
                        current_path = os.path.join(root, 'Current')
                        if os.path.exists(current_path) and not os.path.islink(current_path):
                            os.remove(current_path)
                        os.symlink('A', current_path)

    @staticmethod
    def get_xcode_version():
        '''查询Xcode版本
        
        :returns: str
        '''
        version = "7.3"
        try:
            xcode_info = subprocess.check_output("xcodebuild -version", shell=True, stderr=subprocess.STDOUT)
            if PY3:
                xcode_info = xcode_info.decode()
            version = re.match(r'Xcode\s(\d+\.\d+)', xcode_info.split('\n')[0]).group(1)
        except subprocess.CalledProcessError as e:
            raise Exception('get_xcode_version error:%s' % e.output)
        return version 

    @staticmethod
    def compare_xcode_version(version):
        '''比较当前Mac上的Xcode版本和指定version的大小
        
        当前版本 > version return 1
        当前版本 = version return 0
        当前版本 < version return -1
        '''
        current_version = DT.get_xcode_version()
        return DT.compare_version(current_version, version)

    @staticmethod
    def compare_version(version_a, version_b):
        '''
        version_a > version_b return 1
        version_a = version_b return 0
        version_a < version_b return -1
        '''
        version_reg = r'^\d*\.\d*\.?\d*$'
        if not re.match(version_reg, version_a):
            raise Exception('version_a invalid:%s' % version_a)
        elif not re.match(version_reg, version_b):
            raise Exception('version_b invalid:%s' % version_b)
        a = version_a.split('.')
        b = version_b.split('.')
        length = min(len(a), len(b))
        for i in range(length):
            if int(a[i]) < int(b[i]):
                return -1
            if int(a[i]) > int(b[i]):
                return 1
        if len(a) > len(b):
            return 1
        elif len(a) < len(b):
            return -1
        else:
            return 0

    def get_real_devices(self):
        '''查询当前Mac机上已连接的真机设备（仅含真机，不含模拟器）
        
        :returns: list[dict] or None（意外情况）
        '''
        devices = self.get_devices()
        real_devices = []
        for dev in devices:
            if not dev['simulator']:
                real_devices.append(dev)
        return real_devices
    
    def _get_simulators_below_xcode_7(self):
        devices = []
        try:
            cmd = "xcrun simctl list devices"
            result = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            if PY3:
                result = result.decode()            
            dev_flag = False
            for line in result.split("\n"):
                if line.startswith("-- "):
                    if line.startswith("-- iOS"):
                        dev_flag = True
                        dev_version = re.match(r'--\siOS\s(\d+\.\d)\s--', line).group(1)
                    else:
                        dev_flag = False
                    continue
                if dev_flag:
                    ret = re.match(r'\s{4}(.+)\s\(([0-9A-F\-]+)\)\s\((.+)\)', line)
                    if ret:
                        device = {}
                        device["udid"] = ret.group(2)
                        device["name"] = ret.group(1)
                        device["ios"] = dev_version
                        device["state"] = ret.group(3)
                        device["simulator"] = True
                        devices.append(device)
        except subprocess.CalledProcessError as e:
            raise Exception('_get_simulators_below_xcode_7 error:%s' % e.output)
        return devices
    
    def _get_simulators_above_xcode_7(self):
        devices = []
        try:
            cmd = "xcrun simctl list -j devices"   
            result = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            if PY3:
                result = result.decode()            
            json_devices = json.loads(result, encoding='utf-8')["devices"]
            for k in json_devices:
                if k.startswith("iOS"):
                    ios_version = re.match(r'iOS\s(\d+\.\d+)', k).group(1)
                    for dev in json_devices[k]:
                        dev["ios"] = ios_version
                        dev["simulator"] = True
                        devices.append(dev)
        except subprocess.CalledProcessError as e:
            raise Exception('_get_simulators_above_xcode_7 error:%s' % e.output)
        return devices          
    
    def get_simulators(self):
        '''查询已安装的模拟器(如果没必要查询真机，仅查模拟器的性能更佳)
        
        :returns: list[dict] or None（意外情况）
        '''
        if self.xcode_version < "7.0":
            return self._get_simulators_below_xcode_7()
        else:
            return self._get_simulators_above_xcode_7()
    
    def _list_devices(self):
        '''获取设备列表进程
        '''
        self._list_devices_process = subprocess.Popen("instruments -s devices", 
                                                      shell=True, 
                                                      stdout=subprocess.PIPE, 
                                                      stderr=subprocess.PIPE, 
                                                      close_fds=True)
        self._dev_info, _ = self._list_devices_process.communicate()
    
    def get_devices(self):
        '''查询当前Mac机上的所有可用设备(含真机与模拟器，并产生缓存)
        
        :returns: list[dict]
        '''
        return self._get_device_by_fbsimctl()
    
    def get_device_by_name(self, _name):
        '''通过设备名查询当前Mac机上的可用设备(含真机与模拟器)
        
        :param _name: 设备名(支持正则表达式)，例如：iPhone 5s (8.1 Simulator)
        :type _name: str
        :returns: dict or None
        '''
        _devices = self.get_devices()
        for _item in _devices:
            if _name == _item["name"] or re.match(_name, _item["name"]):
                return _item
    
    def get_device_by_udid(self, _device_udid):
        '''通过设备的udid查询当前Mac机上的设备(含真机与模拟器)
        
        :param _device_udid: 设备的udid
        :type _device_udid: str
        :returns: dict or None
        '''
        _devices = self.get_devices()
        for _item in _devices:
            if _device_udid == _item["udid"]:
                return _item
    
    def is_simulator(self, udid):
        '''判断是否是模拟器
        '''
        device = self.get_device_by_udid(udid)
        if device:
            return device['simulator']
        return False
    
    def check_device_udid_is_valid(self, _device_udid):
        '''检测udid是否有效（含真机与模拟器）
        
        :param _device_udid: 真机或模拟器的udid
        :type _device_udid: str
        :returns: bool
        '''
        device = self.get_device_by_udid(_device_udid)
        return bool(device)
    
    def get_simulator_by_name(self, _name):
        '''通过模拟器名查询模拟器(如果没必要查询真机，仅查模拟器的性能极佳)
        
        :param _name: 模拟器名(支持正则表达式)，例如：iPhone 5s (8.1 Simulator)
        :type _name: str
        :returns: dict or None
        '''
        _simulators = self.get_simulators()
        if _simulators:
            for _item in _simulators:
                if _name == _item["name"] or re.match(_name, _item["name"]):
                    return _item
    
    def get_simulator_by_udid(self, _device_udid):
        '''通过模拟器udid查询模拟器(如果没必要查询真机，仅查模拟器的性能极佳)
        
        :param _device_udid: 模拟器的udid
        :type _device_udid: str
        :returns: dict or None
        '''
        _simulators = self.get_simulators()
        if _simulators:
            for _item in _simulators:
                if _device_udid == _item["udid"]:
                    return _item
    
    def get_default_simulator(self):
        '''查询默认的模拟器(由于模拟器只能有一个实例，所以有默认模拟器。真机不存在默认真机的情况)
        
        :returns: dict or None（意外情况）
        '''
        raise Exception("get_default_simulator unsupported")
    
    def set_default_simulator(self, _device_udid):
        '''设置默认模拟器
        
        :param _device_udid: 模拟器的udid
        :type _device_udid: str
        :returns: bool
        '''
        raise Exception("set_default_simulator unsupported")
    
    def app_info(self, _file_path):
        '''获取APP安装包的信息(IPA或APP均可)
        
        :param _file_path: IPA或APP安装包的路径(注意：当前Mac机必须能访问到该路径)
        :type _file_path: str
        :returns: dict or None
        '''
        raise Exception("app_info unsupported")
    
    def _modify_wi_port(self, udid):
        port = PortManager.get_port('web', udid)
        if self.xcode_version.startswith('9'):
            wi_plist_path = '/Applications/Xcode.app/Contents/Developer/Platforms/iPhoneOS.platform/Developer/Library/CoreSimulator/Profiles/Runtimes/iOS.simruntime/Contents/Resources/RuntimeRoot/System/Library/LaunchDaemons/com.apple.webinspectord.plist'
        else: 
            wi_plist_path = '/Applications/Xcode.app/Contents/Developer/Platforms/iPhoneSimulator.platform/Developer/SDKs/iPhoneSimulator.sdk/System/Library/LaunchDaemons/com.apple.webinspectord.plist'
        cmd = ['/usr/libexec/PlistBuddy', '-c', 'Set :Sockets:com.apple.webinspectord.socket:SockServiceName %s' % port, wi_plist_path]
        subprocess.call(cmd)
    
    def start_simulator(self, udid=None):
        #检查模拟器是否启动成功
        def check_started():
            for s in self.get_simulators():
                if s["state"] == "Booted" and (s["udid"] == udid or udid is None):
                    return True
            return False
        
        if check_started():
            return  True
        simulator = "Simulator"
        if self.compare_xcode_version("7.0") < 0:
            simulator = "iOS Simulator"
        if udid:
            cmd = [self.fbsimctl, udid, 'boot']
            if self.compare_xcode_version("9.0") >= 0:
                cmd = ['xcrun', 'simctl', 'boot', udid]
        else:
            cmd = "open -a \"%s\"" % simulator
        
        if self.compare_xcode_version("9.3") >=0:
            ret = subprocess.call(cmd, close_fds=True)
            if ret != 0:
                result = False
            else:
                result = Timeout(20, 0.05).check(check_started, True)
        else:
            with Flock():
                self._modify_wi_port(udid) 
                ret = subprocess.call(cmd, close_fds=True)
                if ret != 0:
                    result = False
                else:
                    result = Timeout(20, 0.05).check(check_started, True)
        return result
    
    @func_retry_wrap
    def _download_http_file(self, url, target_path):
        '''下载http文件 
        :param url: HTTP路径
        :type url: string
        :param target_path: 下载到目标路径
        :type target_path: string
        '''
        url0 = url
        if url[:7] == 'http://':
            url = url[7:]
        pos = url.find('/')
        host = url[:pos]
        page = url[pos:]
        conn = HTTPConnection(host, port=80, timeout=60)  # 60秒超时
        conn.request('GET', page)
        res = conn.getresponse()
        if res.status != 200: 
            raise RuntimeError('访问：%s 错误[HTTP错误码：%s]' % (url0, res.status))
        data = res.read()
        conn.close()
        f = open(target_path, 'wb')
        f.write(data)
        f.close()
            
    def _download_package(self, src, dstdir):
        '''下载安装包
        
        :param src: 源路径
        :param dstdir: 本地目录
        :returns: filepath - 安装包路径
        '''
        http_prefix = u'http://'
        https_prefix = u'https://'
        
        if src.startswith(http_prefix) or src.startswith(https_prefix):
            '''源路径是http或https服务器路径
            '''
            filename = src[src.rfind('/')+1:]
            filepath = os.path.join(dstdir, filename)
            self._download_http_file(src, filepath)
        
        elif os.path.isfile(src):
            '''源路径是本地路径
            '''
            filepath = src
        
        else:
            raise RuntimeError("file path not supported, %s" % src.encode('utf8'))
        
        return filepath
            
    def _extract_tgz(self, filepath, dstpath):
        '''解压tgz包
        
        :param filepath: tgz包路径
        :type filepath: string
        :param dstpath: 解压后路径
        :type dstpath: string
        '''
        items = os.listdir(dstpath)
        #首先删除目录中已有的app文件，防止版本冲突
        for it in items:
            if it.endswith('.app'):
                shutil.rmtree(os.path.join(dstpath, it))
        zip_decompress(filepath, dstpath)
        items = os.listdir(dstpath)
        for it in items:
            if it.endswith('.app'):
                return os.path.join(dstpath, it)
    
    def _extract_zip(self, filepath, dstpath):
        '''解压zip包
        
        :param filepath: zip包路径
        :type filepath: string
        :param dstpath: 解压后路径
        :type dstpath: string
        '''
        items = os.listdir(dstpath)
        #首先删除目录中已有的app文件，防止版本冲突
        for it in items:
            if it.endswith('.app'):
                shutil.rmtree(os.path.join(dstpath, it))
        zip_decompress(filepath, dstpath)
        items = os.listdir(dstpath)
        for it in items:
            if it.endswith('.app'):
                return os.path.join(dstpath, it)
        
    def _prepare_package(self, _file_path, pkgcachedir):
        '''准备安装包
        
        :param _file_path: 安装包路径
        :type _file_path: string
        :param pkgcachedir: 包缓存目录
        :type pkgcachedir: string
        :returns: str - 安装包本地路径
        '''
        
        if not os.path.isdir(pkgcachedir):
            os.mkdir(pkgcachedir)
        
        if not isinstance(_file_path, six.text_type):
            try:
                _file_path = _file_path.decode('utf8')
            except:
                _file_path = _file_path.decode('gbk')
        
        filepath = self._download_package(_file_path, pkgcachedir)
        if filepath.endswith('.ipa'):
            return filepath
        elif filepath.endswith('.tar.gz'):
            return self._extract_tgz(filepath, pkgcachedir)
        elif filepath.endswith('.zip'):
            return self._extract_zip(filepath, pkgcachedir)
        else:
            raise RuntimeError('unknown type of package: %s' % _file_path.encode('utf8'))

    def _install_for_simulator(self, udid, app_path):
        try:
            cmd = "xcrun simctl install %s %s" % (udid, app_path)
            subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            return (True, "")
        except subprocess.CalledProcessError as e:
            return (False, e.output)

    def install(self, _file_path, _device_udid=None):
        '''通过udid指定真机安装IPA或模拟器安装APP(注意：真机所用IPA不能在模拟器上安装和使用，IPA与APP相互不兼容)
        
        :param _file_path: IPA或APP安装包的路径(注意：当前Mac机必须能访问到该路径)
        :type _file_path: str
        :param _device_udid: 设备的udid(当udid为None时，IPA包则安装到第一台连接上的真机上，APP包则安装到默认模拟器上)
        :type _device_udid: str
        :returns: bool
        '''
        self.install_error = ""
        from qt4i.driver.tools import logger
        log = logger.get_logger("task")
        
        if not _file_path:
            return True
        
        pkgcachedir = os.path.join(os.environ['HOME'],'pkg_cache')
        if not os.path.isdir(pkgcachedir):
            os.mkdir(pkgcachedir)
        pkgcachedir = os.path.join(pkgcachedir, str(_device_udid))
        file_path = self._prepare_package(_file_path, pkgcachedir)
        if os.path.exists(file_path) == False :
            self.install_error = 'file does not exist'
            return False
        if file_path.endswith("ipa"):
            if self.is_simulator(_device_udid):
                log.warn('Failed to install an ipa to a simulator.')
                return False
            ip = InstallationProxy(_device_udid)
            ret = ip.install(file_path)
            
        else:
            if not self.is_simulator(_device_udid):
                log.warn('Failed to install an app to a real device.')
                return False
            self.start_simulator(_device_udid)
            if _device_udid is None:
                _device_udid = "booted"
            ret = self._install_for_simulator(_device_udid, file_path)
        
        if ret[0]:
            if os.path.exists(pkgcachedir):
                shutil.rmtree(pkgcachedir)
        else:
            log.debug('install error: %s' % str(ret[1]))
        
        self.install_error = ret[1]
        return ret[0]
        
    def _uninstall_for_simulator(self, udid, bundle_id):
        try:
            cmd = "xcrun simctl uninstall %s %s" % (udid, bundle_id)
            subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            return (True, "")
        except subprocess.CalledProcessError as e:
            return (False, e.output)
    
    def uninstall(self, _bundle_id, _device_udid=None):
        '''通过udid指定设备卸载指定bundle_id的APP
        
        :param _bundle_id: APP的bundle_id，例如：com.tencent.qq.dailybuild.test
        :type _bundle_id: str
        :param _device_udid: 设备的udid(当udid为None时，如果有真机连接，则卸载第一台连接上的真机内的APP，如果无真机连接，则卸载默认模拟器上的APP)
        :type _device_udid: str
        :returns: bool
        '''
        self.uninstall_error = ""
        simulator_flag = False
        if _device_udid:
            for device in self.get_devices():
                if _device_udid == device["udid"]:
                    simulator_flag = device["simulator"]
                    break
        else:
            devices = self.get_real_devices()
            if len(devices) > 0:
                _device_udid = devices[0]["udid"]
            else:
                devices = self.get_simulators()
                if len(devices) > 0:
                    _device_udid = devices[0]["udid"]
                    simulator_flag = True            

        if _device_udid is None:
            return False
        if simulator_flag:
            self.start_simulator(_device_udid)
            ret = self._uninstall_for_simulator(_device_udid, _bundle_id)
        else:
            ip = InstallationProxy(_device_udid)
            ret = ip.uninstall(_bundle_id)
        
        self.uninstall_error = ret[1]
        return ret[0] 
    
    def _shutdown_simulator(self, udid):
        cmd = [self.fbsimctl, udid, 'shutdown']
        return True if subprocess.call(cmd, close_fds=True) == 0 else False
    
    def reboot(self, udid):
        if self.is_simulator(udid):
            self._shutdown_simulator(udid)
            self.start_simulator(udid)
        else:
            mobiledevice.reboot(udid)

    def get_crash_log(self, udid, procname):
        '''获取指定进程的最新的crash日志
        
        :param udid: 设备的udid
        :type udid: str  
        :param procname: app的进程名，可通过xcode查看
        :type procname: str 
        :return: string or None - crash内容
        '''
        log = None
        if self.is_simulator(udid):
            root_path = os.path.join(os.path.expanduser('~'), 'Library/Logs/DiagnosticReports')
            for crash_log in os.listdir(root_path):
                if crash_log.startswith(procname) and crash_log.endswith('crash'):
                    crash_log_path = os.path.join(root_path, crash_log)
                    with open(crash_log_path, "rb") as fd:
                        log = fd.read().decode("utf-8")
                    if udid in log:
                        os.remove(crash_log_path)
                        return log
                    else:
                        log = None
        else:
            logfile = mobiledevice.get_crash_log(procname, udid)
            if logfile and os.path.isfile(logfile):
                with open(logfile, "rb") as fd:
                    log = fd.read().decode("utf-8")
        return log
    
    def _convert_timestamp(self, format_time):
        parsed_time = time.strptime(format_time, "%b %d %H:%M:%S %Y")
        return time.mktime(parsed_time)
    def _get_simulator_app_data_path(self, udid, bundle_id):
        try:  #先尝试simctl命令获取sandbox路径,如果失败再尝试分析系统安装日志获取路径
            cmd = "xcrun simctl get_app_container %s %s data" % (udid, bundle_id)
            app_path = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT, close_fds=True)
            if PY3:
                app_path = app_path.decode()            
            return app_path[:-1]
        except subprocess.CalledProcessError:
            pass
        install_log_path = os.path.join(os.path.expanduser('~'), 'Library/Logs/CoreSimulator/%s/MobileInstallation' % udid)
        apps_path = os.path.join(os.path.expanduser('~'), 'Library/Developer/CoreSimulator/Devices/%s/data/Containers/Data/Application' % udid)
        app_path = None
        timestamp = 0
        for logfile in os.listdir(install_log_path):
            if logfile.startswith('mobile_installation.log'):
                with open(os.path.join(install_log_path, logfile), "rb") as fd:
                    for line in fd:
                        if 'Made container live for %s at %s' % (bundle_id, apps_path) in line :
                            line_timestamp = self._convert_timestamp(line[4:24])
                            if line_timestamp > timestamp:
                                app_path = line
                                timestamp = line_timestamp
                        elif 'Data container for %s is now at %s' % (bundle_id, apps_path) in line:
                            line_timestamp = self._convert_timestamp(line[4:24])
                            if line_timestamp > timestamp:
                                app_path = line
                                timestamp = line_timestamp
        return app_path[app_path.find(apps_path):-1]
    
    def push_file(self, udid, bundle_id, localpath, remotepath):
        '''拷贝Mac本地文件到iOS设备sandbox的指定目录中
        
        :param udid: 设备的udid,如果不传入udid，则表示向手机的系统目录传入文件
        :type udid: str 
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param localpath: Mac上的文件路径
        :type localpath: str
        :param remotepath: sandbox上的目录，例如：/Library/Caches/test/
        :type remotepath: str
        :returns: boolean
        '''
        if bundle_id:
            if self.is_simulator(udid):
                sandbox_root = self._get_simulator_app_data_path(udid, bundle_id)
                remotepath = remotepath[1:] if remotepath.startswith('/') else remotepath
                remote_file = os.path.join(sandbox_root, remotepath)
                shutil.copy(localpath, remote_file)
                return True
            else:
                return mobiledevice.push_file(bundle_id, localpath, remotepath, udid)
        else:
            if self.is_simulator(udid):
                cmd = 'xcrun simctl addmedia %s %s' % (udid, localpath)
                process = subprocess.Popen(cmd, shell=True, stderr=subprocess.STDOUT)
                
                process.communicate()
                return process.poll() == 0
            else:
                raise Exception("This operation is not supported by the device")
 
    def pull_file(self, udid, bundle_id, remotepath, localpath='/tmp', is_dir=False, is_delete = True):
        '''拷贝手机中sandbox指定目录的文件到Mac本地
        
        :param udid: 设备的udid
        :type udid: str 
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param remotepath: sandbox上的目录或者文件，例如：/Library/Caches/test/
        :type remotepath: str        
        :param localpath: 本地的目录
        :type localpath: str
        :param is_dir: remotepath是否为目录，默认为单个文件
        :type is_dir: bool
        :param is_delete: 是否删除
        :type is_delete: bool
        :returns: list or None
        '''
        if self.is_simulator(udid):
            local_files = []
            sandbox_root = self._get_simulator_app_data_path(udid, bundle_id)
            remotepath = remotepath[1:] if remotepath.startswith('/') else remotepath
            remote_file = os.path.join(sandbox_root, remotepath)
            if is_dir:
                for root, _, files in os.walk(remote_file):
                    for f in files:
                        if udid:
                            (short_name, extension) = os.path.splitext(f)
                            new_f = '%s-%s%s' % (short_name, udid, extension)
                        filename = os.path.join(root, f)
                        local_file = os.path.join(localpath, new_f)
                        shutil.copy(filename, local_file)
                        local_files.append(local_file)
                        if is_delete:
                            os.remove(filename)
            else:
                if udid:
                    (short_name, extension) = os.path.splitext(os.path.basename(remotepath))
                    new_f = '%s-%s%s' % (short_name, udid, extension)
                local_file = os.path.join(localpath, new_f)
                shutil.copy(remote_file, local_file)
                local_files.append(local_file)
                if is_delete:
                    os.remove(remote_file)
            return local_files
        else:
            return mobiledevice.pull_file(bundle_id, remotepath, localpath, udid, is_dir, is_delete)
 
    def remove_files(self, udid, bundle_id, file_path):
        '''删除手机上指定app中的文件或者目录
        
        :param udid: 设备的udid
        :type udid: str 
        :param bundle_id: app的bundle id
        :type bundle_id: str 
        :param file_path: 待删除的文件或者目录，例如: /Documents/test.log
        :type file_path: str  
        '''
        if self.is_simulator(udid):
            sandbox_root = self._get_simulator_app_data_path(udid, bundle_id)
            file_path = file_path[1:] if file_path.startswith('/') else file_path
            file_path = os.path.join(sandbox_root, file_path)
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.exists(file_path):
                shutil.rmtree(file_path)
        else:
            mobiledevice.remove_files(bundle_id, file_path, udid)
            
    def list_apps_with_fbsimctl(self, udid, app_type):
        try:
            apps_info =[]
            cmd = "%s --json %s list_apps" % (self.fbsimctl, udid)
            result = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            if PY3:
                result = result.decode()             
            result = result.split('\n')
            for line in result:
                if 'list_apps' in line:
                    line = json.loads(line, encoding='utf-8')
                    if line['event_type'] == 'discrete':
                        apps_info = line['subject']
                        break
            apps = []
            for app_info in apps_info:
                if app_type == 'all':
                    apps.append({app_info['bundle']['bundle_id']:app_info['bundle']['name']})
                elif app_info['install_type'] == app_type:
                    apps.append({app_info['bundle']['bundle_id']:app_info['bundle']['name']})
            return apps
        except subprocess.CalledProcessError as e:
            raise Exception('list_apps for simulator error:%s' % e.output)
    
    def list_apps(self, udid, app_type): 
        '''获取设备上的app列表
        
        :param udid: 设备的udid
        :type udid: str 
        :param app_type: app的类型(user/system/all)
        :type app_type: str 
        :returns: list
        '''
        if self.is_simulator(udid):
            return self.list_apps_with_fbsimctl(udid, app_type)
        else:
            return InstallationProxy(udid).list_apps(app_type)
    
    def get_syslog(self, watchtime, logFile, processName, udid):
        '''获取设备上服务日志

        :param watchtime: 观察时间
        :type watchtime: int
        :param logFile: 日志文件名
        :type logFile: str
        :param processName: 服务名
        :type processName: str
        :param udid: 设备的udid
        :type udid: str
        '''
        if self.is_simulator(udid):
            root_path = os.path.join(os.path.expanduser('~'), 'Library/Logs/CoreSimulator/')
            syslog_path = os.path.join(os.path.join(root_path, udid),'system.log')
            with open(syslog_path,"rb") as fd:
                lines = fd.readlines()
            pattern = re.compile(r'\d{2}:\d{2}:\d{2}')
            time_end = pattern.search(lines[-1]).group()
            timestamp = int(time_end[0:2])*3600 + int(time_end[3:5])*60 + int(time_end[6:8])
            a = datetime.datetime.utcfromtimestamp(timestamp) 
            b = datetime.timedelta(seconds=watchtime)  
            start = (str(a-b))[-8:]
            time_start = int(start[6:8])+60*int(start[3:5])+3600*int(start[0:2])
            log = []
            for i in range(len(lines)):
                log.append(lines[len(lines)-i-1])
                if pattern.search(lines[len(lines)-i-1]):
                    tm = pattern.search(lines[len(lines)-i-1]).group()
                    time_now = int(tm[6:8])+60*int(tm[3:5])+3600*int(tm[0:2]) 
                    if (time_now < time_start):
                        break
                    
            with open(logFile, 'a') as f:
                for i in log:
                    f.write(i.replace("\x00", ""))       
        else:
            mobiledevice.get_syslog(watchtime, logFile, processName, udid)
    
    def update_sandbox_client(self, udid, bundle_id):
        '''获取SandboxClient对象，防止一直创建对象导致StartServiceError
        
        :param udid: 设备的udid
        :type udid: str
        :param bundle_id: 应用的bundle_id
        :type bundle_id: str
        '''
        if not self.udid or not self.bundle_id:
            self.udid = udid
            self.bundle_id = bundle_id
            self.sc = SandboxClient(udid, bundle_id)
        elif self.udid != udid or self.bundle_id != bundle_id:
            self.close_sandbox_client()
            self.sc = SandboxClient(udid, bundle_id)
        else:
            try:
                #调用sandboxClient接口方法，以确认该对象是否还可用
                self.sc.read_directory('/')
            except:
                self.sc = SandboxClient(udid, bundle_id)
    
    def close_sandbox_client(self):
        '''对象销毁
        '''
        self.bundle_id = None
        self.udid = None
        self.sc.close()
    
    def get_sandbox_path_files(self, udid, bundle_id, file_path):
        '''返回真机或者模拟器的沙盒路径
        
        :param udid: 设备的udid
        :type udid: str
        :param bundle_id: 应用的bundle_id
        :type bundle_id: str
        :param file_path: 沙盒目录
        :type file_path: str
        '''
        file_path = file_path.encode('utf-8')
        sandbox_tree = []
        tmp_dict = {}
        if self.is_simulator(udid):
            sandbox_root = self._get_simulator_app_data_path(udid, bundle_id)
            file_path = sandbox_root if file_path == '/' else file_path
            for l in os.listdir(file_path):
                tmp_dict['path'] = os.path.join(file_path, l)
                tmp_dict['is_dir'] = os.path.isdir(tmp_dict['path'])
                sandbox_tree.append(tmp_dict)
                tmp_dict = {}
        else:
            self.update_sandbox_client(udid, bundle_id)
            for l in self.sc.read_directory(file_path):
                if l not in ('.', '..'):
                    tmp_dict['path'] = os.path.join(file_path, l)
                    info = self.sc.get_file_info(tmp_dict['path'])
                    tmp_dict['is_dir'] = (info != None and info['st_ifmt'] == 'S_IFDIR')
                    sandbox_tree.append(tmp_dict)
                    tmp_dict = {}
        return sandbox_tree
     
    def is_sandbox_path_dir(self, udid, bundle_id, file_path):
        '''判断一个sandbox路径是否是一个目录
        
        :param udid: 设备的udid
        :type udid: str
        :param bundle_id: 应用的bundle_id
        :type bundle_id: str
        :param file_path: 沙盒目录
        :type file_path: str
        '''
        file_path = file_path.encode('utf-8')
        if self.is_simulator(udid):
            sandbox_root = self._get_simulator_app_data_path(udid, bundle_id)
            file_path = sandbox_root if file_path == '/' else file_path
            return os.path.isdir(file_path)
        else:
            self.update_sandbox_client(udid, bundle_id)
            info = self.sc.get_file_info(file_path)
            return (info != None and info['st_ifmt'] == 'S_IFDIR')
      
    def get_sandbox_file_content(self, udid, bundle_id, file_path):
        '''获取sandbox中文本文件的内容
        
        :param udid: 设备的udid
        :type udid: str
        :param bundle_id: 应用的bundle_id
        :type bundle_id: str
        :param file_path: 沙盒目录
        :type file_path: str
        '''
        file_path = file_path.encode('utf-8')
        if self.is_simulator(udid):
            if os.path.exists(file_path):
                with open(file_path, "rb") as fd:
                    return base64.b64encode(fd.read())
            else:
                raise Exception('file(%s) does not exist' % file_path)
        else:
            self.update_sandbox_client(udid, bundle_id)
            content = self.sc.get_file_contents(file_path)
            if content:
                return base64.b64encode(content) 
            
    def _get_device_by_fbsimctl(self):
        '''获取设备列表
        '''
        devices = []
        cmd = "%s --json list" % self.fbsimctl
        try:
            res = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            if PY3:
                res = res.decode()
            for dev in res.split('\n'):
                if dev == '' :
                    continue
                dev = json.loads(dev)['subject']
                if 'name' not in dev:
                    continue
                name = dev['name']
                if 'Apple Watch' in name:
                    continue
                if 'Apple TV' in name:
                    continue
                arch = dev['arch']
                if arch.startswith('arm64'):
                    dev[u'simulator'] = False
                else :
                    dev[u'simulator'] = True 
                ios = dev['os']
                dev[u'ios'] = ios.split(' ')[1].strip()
                devices.append(dev)
        except subprocess.CalledProcessError:
            pass
        return devices



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''RPC Framework
'''

from __future__ import absolute_import, print_function

import json
import random
import re
import string
import six
from six.moves.socketserver import ThreadingMixIn
from six.moves.xmlrpc_server import SimpleXMLRPCServer
from six.moves.xmlrpc_server import SimpleXMLRPCRequestHandler
import six.moves.xmlrpc_client as xmlrpc_client
from six.moves.http_client import HTTPResponse
from six import StringIO
from six import PY2

if PY2: 
    from urllib import splittype
    from urllib import splithost
else:
    from urllib.parse import splittype
    from urllib.parse import splithost

from qt4i.driver.tools import logger


try:
    import fcntl
except ImportError:
    fcntl = None

IDCHARS = string.ascii_lowercase+string.digits


def random_id(length=8):
    return_id = ''
    for _ in range(length):
        return_id += random.choice(IDCHARS)
    return return_id


class _RPCMethod(object):
    """RPC method decorator
    """
    def __init__(self, method, instance=None, owner=None ):
        self.method = method
        self.instance = instance
        self.owner = owner
         
    def __call__(self, *args, **kwargs ):
        if self.instance == None or self.owner == None:
            return self.method(*args, **kwargs)
        else:
            return self.method.__get__(self.instance, self.owner)(*args, **kwargs)
    
    def __get__(self, instance, owner):
        return _RPCMethod(self.method, instance, owner)

rpc_method = _RPCMethod


class RPCEndpoint(object):
    """RPC end point, subclass to define a RPC end point
    """
    rpc_name_prefix = ""
    
    def _dispatch(self, method, params):
        if not method.startswith(self.rpc_name_prefix):
            raise Exception('method "%s" is not supported by endpoint "%s"' % (method, type(self).__name__))
        try:
            m = getattr(self, method[len(self.rpc_name_prefix):])
        except AttributeError:
            raise Exception('method "%s" is not supported by endpoint "%s"' % (method, type(self).__name__))
        else:
            if not isinstance(m, _RPCMethod):
                raise Exception('method "%s" is not exposed by endpoint "%s"' % (method, type(self).__name__))
            return m(*params)
        
    def create_json_command(self, method, *params):
        '''serialize function to JSON Object
        '''
        return {"method": method, "params": list(params)}


class FakeSocket():
    def __init__(self, response_str):
        self._file = StringIO(response_str)
    def makefile(self, *args, **kwargs):
        return self._file


class RPCClientProxy(object):
    '''RPC Client, 
    copy  ServerProxy 的源码， 在__request 方法中进行解码
    主要实现RPC客户端返回Unicode转换为UTF-8
    '''

    def __init__(self, uri, ws_uri=None,transport=None, encoding=None, verbose=0,
                 allow_none=0, use_datetime=0, context=None):
        # establish a "logical" server connection
        if PY2 and isinstance(uri, six.text_type):
            uri = uri.encode('ISO-8859-1')
        # get the url
        protocol, uri = splittype(uri)
        if protocol not in ("http", "https"):
            raise IOError("unsupported JSON-RPC protocol")
        self.__host, self.__handler = splithost(uri)
        if not self.__handler:
            self.__handler = "/RPC2"
        self.__ws_uri = ws_uri

        if transport is None:
            if protocol == "https":
                transport = SafeTransport(use_datetime=use_datetime, context=context)
            else:
                transport = Transport(use_datetime=use_datetime)
        self.__transport = transport

        self.__encoding = encoding
        self.__verbose = verbose
        self.__allow_none = allow_none

    def __close(self):
        self.__transport.close()

    def _parse_ws_response(self):
        http_response = self.__ws.recv()
        response = HTTPResponse(FakeSocket(http_response))
        response.begin()
        while response.length > len(http_response):
            http_response += self.__ws.recv()
        response = HTTPResponse(FakeSocket(http_response))
        self.__transport.verbose = 0
        response.begin()
        return self.__transport.parse_response(response)

    def _ws_request(self, request):
        import websocket
        self.__ws = websocket.WebSocket()
        self.__ws.connect(self.__ws_uri)
        header = "POST {abs_path} HTTP/1.1\r\n" \
                "Host: {host}\r\n"  \
                "User-Agent: jsonrpclib/0.1\r\n" \
                "Connection: close\r\n"  \
                "Accept-Encoding: gzip\r\n" \
                "Content-Type: application/json-rpc\r\n"  \
                "Content-Length: {len}\r\n" \
                "\r\n".format(abs_path=self.__handler, host=self.__host, len=len(request)).encode("UTF-8")
        self.__ws.send(header + request)
        return self._parse_ws_response()

    def __request(self, methodname, params):
        # call a method on the remote server
        request = {"jsonrpc": "2.0"}
        if len(params) > 0:
            request["params"] = params
        request["id"] = random_id()
        request["method"] = methodname
        request = json.dumps(request)
        if self.__ws_uri:
            response = self._ws_request(request)
        else:
            response = self.__transport.request(
                self.__host,
                self.__handler,
                request,
                verbose=self.__verbose
                )
        response = json.loads(response)
        if not isinstance(response, dict):
            raise TypeError('Response is not dict')
        if 'error' in response.keys() and response['error'] is not None:
            raise DriverApiError(response['error']['message'])
        else:
            response = response['result'][0]
            if isinstance(response, dict):
                return self.encode_dict(response, "UTF-8")
            elif isinstance(response, list):
                return self.encode_list(response, "UTF-8")
            elif PY2 and isinstance(response, six.text_type):
                return response.encode("UTF-8")
            return response

    def __repr__(self):
        return (
            "<ServerProxy for %s%s>" %
            (self.__host, self.__handler)
            )

    __str__ = __repr__

    def __getattr__(self, name):
        # magic method dispatcher
        return xmlrpc_client._Method(self.__request, name)

    # note: to call a remote object with an non-standard name, use
    # result getattr(server, "strange-python-name")(args)

    def __call__(self, attr):
        """A workaround to get special attributes on the ServerProxy
           without interfering with the magic __getattr__
        """
        if attr == "close":
            return self.__close
        elif attr == "transport":
            return self.__transport
        raise AttributeError("Attribute %r not found" % (attr,))
    
    def encode_dict(self, content, encoding="UTF-8"):
        '''将字典编码为指定形式
        
        :param content: 要编码内容
        :type content: dict
        :param encoding:编码类型
        :type encoding: str
        :returns: dict -- 编码后的字典
        '''
        for key in content:
            if isinstance(content[key], dict):
                content[key] = self.encode_dict(content[key], encoding)
            elif PY2 and isinstance(content[key], six.text_type):
                content[key] = content[key].encode(encoding)
            elif isinstance(content[key], list):
                content[key] = self.encode_list(content[key], encoding)
        return content   
    
    def encode_list(self, content, encoding="UTF-8"):
        '''将列表编码为指定形式
        
        :param content: 要编码内容
        :type content: list
        :param encoding:编码类型
        :type encoding: str
        :returns: list -- 编码后的列表
        '''
        for ind, item in enumerate(content):
            if isinstance(item, dict):
                content[ind] = self.encode_dict(item, encoding)
            elif PY2 and isinstance(item, six.text_type):
                content[ind] = content[ind].encode(encoding)
            elif isinstance(item, list):
                content[ind] = self.encode_list(item, encoding)
        return content


class DriverApiError(Exception):
    '''Driver API Error
    '''


class Fault(object):
    '''JSON-RPC Error
    '''

    def __init__(self, code=-12306, message = None, rpcid=None):
        self.faultCode = code
        self.faultString = message
        self.rpcid = rpcid
        if not message:
            import traceback
            self.faultString = traceback.format_exc()

    def error(self):
        return {"code": self.faultCode, "message": self.faultString}
    
    def response(self):
        return json.dumps({"jsonrpc": "2.0", "error":self.error(), "id":self.rpcid})

    def __repr__(self):
        return '<Fault %s: %s>' % (self.faultCode, self.faultString)

 
class TransportMixIn(object):
    '''XMLRPC Transport extended API
    '''
    user_agent = "jsonrpclib/0.1"
    _connection = (None, None)
    _extra_headers = []
        
    def send_content(self, connection, request_body):
        connection.putheader("Content-Type", "application/json-rpc")
        connection.putheader("Content-Length", str(len(request_body)))
        connection.endheaders()
        if request_body:

            if not PY2 and isinstance(request_body, str):
                request_body = request_body.encode('utf-8')
            connection.send(request_body)
            
    def getparser(self):
        target = JSONTarget()
        return JSONParser(target), target    
    

class JSONParser(object):
    
    def __init__(self, target):
        self.target = target

    def feed(self, data):
        self.target.feed(data)

    def close(self):
        pass


class JSONTarget(object):
    
    def __init__(self):
        self.data = []

    def feed(self, data):
        self.data.append(data)

    def close(self):
        if PY2:
            return ''.join(self.data)
        else:
            return b''.join(self.data)


class Transport(TransportMixIn, xmlrpc_client.Transport):
    
    def __init__(self, use_datetime):
        TransportMixIn.__init__(self)
        xmlrpc_client.Transport.__init__(self, use_datetime)


class SafeTransport(TransportMixIn, xmlrpc_client.SafeTransport):
    
    def __init__(self, use_datetime, context):
        TransportMixIn.__init__(self)    
        xmlrpc_client.SafeTransport.__init__(self, use_datetime, context)


class SimpleJSONRPCRequestHandler(SimpleXMLRPCRequestHandler):
    '''JSON-RPC请求处理器
    '''
    def is_rpc_path_valid(self):
        return True
    
    def do_POST(self):
        '''处理HTTP的POST请求
        '''
        if not self.is_rpc_path_valid():
            self.report_404()
            return
        try:
            max_chunk_size = 10*1024*1024
            size_remaining = int(self.headers["content-length"])
            L = []
            while size_remaining:
                chunk_size = min(size_remaining, max_chunk_size)
                L.append(self.rfile.read(chunk_size))
                size_remaining -= len(L[-1])
            if PY2:
                data = ''.join(L)
            else:
                data = b''.join(L)
            response = self.server._marshaled_dispatch(
                    data, getattr(self, '_dispatch', None), self.path
                )
            self.send_response(200)
        except Exception:
            response = Fault().response()
            self.send_response(500, response)
            logger.get_logger().exception("ProtocolError:%s" % response)
        if response is None:
            response = ''
        self.send_header("Content-Type", "application/json-rpc")
        if self.encode_threshold is not None:
            if len(response) > self.encode_threshold:
                    q = self.accept_encodings().get("gzip", 0)
                    if q:
                        try:
                            response = xmlrpc_client.gzip_encode(response)
                            self.send_header("Content-Encoding", "gzip")
                        except NotImplementedError:
                            pass
        self.send_header("Content-length", str(len(response)))
        self.end_headers()
        if PY2:
            self.wfile.write(response)
        else:
            self.wfile.write(response.encode('utf-8'))


class SimpleJSONRPCServer(ThreadingMixIn, SimpleXMLRPCServer):
    """RPC Server
    """
    
    LOG_FILTERED_METHODS = [
                            'push_file_data', 
                            'pull_file_data',
                            'device.get_log',
                            'device.get_crash_log',
                            'device.get_driver_log',
                            'device.get_element_tree', 
                            'device.get_element_tree_and_capture_screen',
                            'device.capture_screen', 
                            'element.get_element_tree',
                            ]

    def __init__(self, urls, addr):
        """Constructor
        
        :param urls: list of URL pattern and RPC end point class
        :param addr: listening address
        :type urls: list
        :type addr: tuple
        """
        SimpleXMLRPCServer.__init__(self, addr=addr, 
                                                       requestHandler=SimpleJSONRPCRequestHandler,
                                                       allow_none=True, 
                                                       encoding='UTF-8',
                                                       logRequests=False)
        
        self._dispatcher_patterns = []
        for url_pattern, method_pattern, endpoint_cls in urls:
            if method_pattern:
                method_c_pattern = re.compile(method_pattern)
            else:
                method_c_pattern = None
            self._dispatcher_patterns.append((re.compile(url_pattern),
                                              method_c_pattern,
                                              endpoint_cls))
    
    def _marshaled_dispatch(self, data, dispatch_method=None, path=None):
        origin_path = path
        path = path[1:] #remove /
        if not path.endswith('/'):
            path += '/'
        
        try:
            request = json.loads(data)
        except ValueError:
            fault = Fault(-32700, 'JSON parsing error')
            return fault.response()
        if not isinstance(request, dict):
            fault = Fault(-32600, 'Invalid request data type')
            return fault.response()
        rpcid = request.get('id', None)
        method = request.get('method', None)
        params = request.get('params', [])
        params_types = (list, dict, tuple)
        if not method or not isinstance(method, six.string_types) or not isinstance(params, params_types):
            return Fault(-32600, 'Invalid request method or parameters').response()
          
        tried_enpoint_clss = []
        for url_pattern, method_pattern, endpoint_cls in self._dispatcher_patterns:
            m = url_pattern.match(path)
            if m:
                try:
                    log_suffix = m.group(1)
                    log = logger.get_logger("driverserver_%s" % log_suffix)
                except:
                    log = logger.get_logger()
                
                tried_enpoint_clss.append(endpoint_cls)
                if (method_pattern is None) or method_pattern.match(method):
                    endpoint_params = m.groupdict()
                    try:
                        endpoint = endpoint_cls(self, **endpoint_params)
                    except:
                        fault = Fault()
                        log.exception(fault.error())
                        return fault.response()
                    else:    
                        break
        else:
            if not tried_enpoint_clss:
                return Fault(-32601, "invalid URL: \"%s\"" % origin_path).response()
            else:
                return Fault(-32601, "invalid method: \"%s\", no matched end point, end point: %s tried" % 
                    (method, ", ".join([ '"%s"'%it.__name__ for it in tried_enpoint_clss]))).response()
          
        try:
            log.debug('--- --- --- --- --- ---')
            log.debug('%s <<< %s' % (method, params))
            
            response = endpoint._dispatch(method, params)
            if method in self.LOG_FILTERED_METHODS:
                log.debug('%s >>> done' % method) 
            else:
                log.debug('%s >>> %s' % (method, response))            
            # wrap response in a singleton tuple
            response = (response,)
            response = json.dumps({"jsonrpc": "2.0", "result": response, "id": rpcid})
        except:
            fault = Fault()
            response = fault.response()
            log.error('%s >>> %s' % (method, fault.error()))
            log.exception(fault.error())
        return response


# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''pymobiledevice extended tools
'''

from __future__ import absolute_import, print_function

from cmd import Cmd
import os
import re
from subprocess import check_call, STDOUT, CalledProcessError
import tempfile
import time
import six

from pymobiledevice.afc import AFCClient, AFCShell
from pymobiledevice.diagnostics_relay import DIAGClient
from pymobiledevice.lockdown import LockdownClient
from pymobiledevice.screenshotr import screenshotr
from pymobiledevice.usbmux.usbmux import USBMux
from pymobiledevice.syslog import Syslog


TIME_FORMAT = '%Y-%m-%d %H:%M:%S'

def list_devices():
    '''获取当前电脑上连接的所有iOS设备
    
    :returns: list - 包含所有iOS设备的UDID的数组
    '''
    mux = USBMux()
    if not mux.devices:
        mux.process(1)
    for _ in range(5): 
        mux.process(0.1)
    return [d.serial for d in mux.devices]


def get_device_info(udid = None, keyname = None):
    '''获取iOS设备的相关属性信息
    
    :param udid: iOS设备的udid
    :type udid: str
    :param keyname: iOS设备的特定信息字段的名称
    :type keyname: str
    :returns: str - iOS设备的信息
    ''' 
    lockdown = LockdownClient(udid)
    lockdown.startService("com.apple.afc")
    if keyname:
        return lockdown.allValues.get(keyname)
    else:
        return lockdown.allValues

    
def get_screenshot(udid = None, filepath = "/tmp/screenshot.png"):  
    '''获取iOS设备的屏幕快照
    
    :param udid: iOS设备的udid
    :type udid: str
    :param filepath: 屏幕快照的存储路径
    :type filepath: str
    :returns: boolean - 截图是否成功
    '''   
    result = False
    tiff_file = tempfile.NamedTemporaryFile(suffix='.tiff')
    tiff_file_path = tiff_file.name
    lockdown = LockdownClient(udid)
    screenshot = screenshotr(lockdown)
    data = screenshot.take_screenshot()
    with open(tiff_file_path, "wb") as fd:
        fd.write(data)
    screenshot.stop_session()
    try:
        args = ["/usr/bin/sips", "-s format png", tiff_file_path, "--out", filepath]
        check_call(" ".join(args), shell=True, stderr=STDOUT)
        result = True
    except CalledProcessError:
        pass
    finally:
        tiff_file.close()
    return result


def get_crash_log(procname, device_udid = None, log_path='/tmp'):
    afc_crash = AFCCrashLog(device_udid)
    return afc_crash.get_crash_log(procname, log_path, True)


def pull_file(bundle_id, remotepath, localpath='/tmp', device_udid = None, is_dir=False, is_delete = True):
    '''拷贝手机中sandbox指定目录的文件到Mac本地
    
    :param bundle_id: app的bundle id
    :type bundle_id: str   
    :param remotepath: sandbox上的目录或者文件，例如：/Library/Caches/test/
    :type remotepath: str        
    :param localpath: 本地的目录
    :type localpath: str
    :param device_udid: 设备的udid
    :type device_udid: str
    :param is_dir: remotepath是否为目录，默认为单个文件
    :type is_dir: bool
    :param is_delete: 是否删除
    :type is_delete: bool
    :returns: list or None
    '''
    afc_client = None
    try:
        files = None
        afc_client = SandboxClient(device_udid, bundle_id)
        files = afc_client.copy_to_local(remotepath, localpath, is_dir, is_delete)
    finally:
        if afc_client is not None:
            afc_client.close()
    return files


def push_file(bundle_id, localpath, remotepath, device_udid = None):
    '''拷贝Mac本地文件到手机中sandbox的指定目录地
    
    :param bundle_id: app的bundle id
    :type bundle_id: str   
    :param localpath: Mac上的文件路径
    :type localpath: str
    :param remotepath: sandbox上的目录或者文件，例如：/Library/Caches/test/
    :type remotepath: str        
    :param device_udid: 设备的udid
    :type device_udid: str
    :returns: boolean
    '''
    afc_client = None
    result = False
    try:
        afc_client = SandboxClient(device_udid, bundle_id)
        result = afc_client.copy_to_remote(localpath, remotepath)
        result = True
    finally:
        if afc_client is not None:
            afc_client.close()
    return result


def remove_files(bundle_id, file_path, device_udid = None):
    '''删除手机上指定app中的文件或者目录
    
    :param bundle_id: app的bundle id
    :type bundle_id: str 
    :param file_path: 待删除的文件或者目录，例如: /Documents/test.log
    :type file_path: str  
    :param device_udid: 设备的udid
    :type device_udid: str
    '''
    afc_client = None
    try:
        afc_client = SandboxClient(device_udid, bundle_id)
        afc_client.remove(file_path)
    finally:
        if afc_client is not None:
            afc_client.close()

    
def reboot(device_udid = None):
    '''重启手机
    
    :param device_udid: 设备的udid
    :type device_udid: str
    '''
    lockdown = LockdownClient(udid=device_udid)
    DIAGClient(lockdown).restart()
    time.sleep(10)
    for _ in range(20):
        if device_udid in list_devices():
            print('reboot successfully')
            break
    else:
        print('reboot error: real device disconect')


def get_syslog(watchtime, logFile, procName = None, device_udid = None):
    ''' 获取手机服务日志
    
    :param watchtime: 观察时间
    :type watchtime: 观察时间
    :param logFile: 日志名称
    :type device_udid: str
    :param procName: 进程名称
    :type procName: str
    :param device_udid: 设备的udid
    :type device_udid: str
    '''
    lockdown = LockdownClient(udid=device_udid)
    syslog = Syslog(lockdown)
    syslog.watch(watchtime, logFile, procName)
              

class InstallationProxy(object):
    
    def __init__(self, udid=None):
        self.udid = udid
        self.lockdown = LockdownClient(udid)
        self.service = self.lockdown.startService("com.apple.mobile.installation_proxy")
    
    def wait_completion(self, handler=None, *args):
        while True:
            z =  self.service.recvPlist()
            if not z:
                break
            completion = z.get("PercentComplete")
            if completion:
                if handler:
                    handler(completion,*args)    
#                 print "%s: %s%% Complete" % (z.get("Status"), completion)
            else:
                if z.get("Status") == "Complete" or ("Status" not in z and "CFBundleIdentifier" in z):
                    return (True, "")
                else:
                    return (False, z.get("ErrorDescription"))
    
    def send_cmd_for_bid(self, bundle_id, cmd="Archive", options=None, handler=None, *args):
        cmd = {"Command": cmd, "ApplicationIdentifier": bundle_id }
        if options:
            cmd.update(options) 
        self.service.sendPlist(cmd)
        return self.wait_completion(handler, *args)      
        
    def install(self, ipa_path, options=None, handler=None, *args):
        '''安装应用程序
        
        :param ipa_path: 安装包的路径
        :type ipa_path: str
        :return: boolean - 安装是否成功
        '''
        print("上传安装包...")
        afc_client = AFCClient(self.lockdown)
        tmp_ipa = "t%d.ipa" % time.time()
        with open(ipa_path, "rb") as f:
            ipa_content = f.read()
            afc_client.set_file_contents("/" + tmp_ipa, ipa_content)
            print("上传完毕")
        print("开始安装")
        cmd = {"Command":"Install", "PackagePath": tmp_ipa}
        if options:
            cmd.update(options)
        self.lockdown = LockdownClient(self.udid)
        self.service = self.lockdown.startService("com.apple.mobile.installation_proxy")
        self.service.sendPlist(cmd)
        ret = self.wait_completion(handler, args)
        return ret

    
    def uninstall(self, bundle_id, options=None, handler=None, *args):
        '''卸载应用程序
    
        :param bundle_id: 应用程序的bundle id
        :type bundle_id: str
        :return: boolean - 卸载是否成功
        '''     
        return self.send_cmd_for_bid(bundle_id, "Uninstall", options, handler, args)
    
    def apps_info(self):
        self.service.sendPlist({"Command": "Lookup"})
        return self.service.recvPlist()
    
    def list_apps(self, app_type='all'):
        options = {}
        if app_type == 'system':
            options["ApplicationType"]="System"
        elif app_type == 'user':
            options["ApplicationType"]="User"
        options["ReturnAttributes"] = ["CFBundleIdentifier",
#                                        "CFBundleDisplayName",
#                                        "CFBundleVersion",
                                        "CFBundleName",]
        self.service.sendPlist({"Command": "Browse", "ClientOptions":options})
        apps = []
        while True:
            z = self.service.recvPlist()
            if z.get("Status") == "BrowsingApplications":
                apps.extend(z["CurrentList"])
            elif z.get("Status") == "Complete":
                break
            else:
                raise Exception(z.get("ErrorDescription"))
        apps = [{app["CFBundleIdentifier"]:app["CFBundleName"]} for app in apps]
        return apps
    
    def __del__(self):
        if hasattr(self, "service"):
            self.service.close()
        
        
class AFCShell2(AFCShell):
    
    def __init__(self, afcclient, completekey='tab', stdin=None, stdout=None):
        Cmd.__init__(self, completekey=completekey, stdin=stdin, stdout=stdout)
        self.afc = afcclient
 
        self.curdir = '/'
        self.prompt = 'AFC$ ' + self.curdir + ' '
        self.complete_cat = self._complete
        self.complete_ls = self._complete
    
    def dir_walk(self,dirname):
        res = self.afc.dir_walk(dirname)
        file_iter = six.next(res)[2]
        return file_iter
    
    def set_file_contents(self,filename,data):
        return self.afc.set_file_contents(filename,data)

    def get_file_contents(self,filename):
        return self.afc.get_file_contents(filename)
    
    def do_rm(self, p):
        f =  self.afc.get_file_info(self.curdir + "/" + p)
        if f is None:
            return  
        elif 'st_ifmt' in f and f['st_ifmt'] == 'S_IFDIR':
            self.afc.remove_directory(self.curdir + "/" + p)
        else:
            self.afc.file_remove(self.curdir + "/" + p)    


class AFCCrashLog(AFCClient):
    def __init__(self, udid=None, lockdown=None):
        self.lockdown = lockdown if lockdown else LockdownClient(udid)
        self.service = None
        retry = 5
        while retry > 0:
            try:
                self.service = self.lockdown.startService("com.apple.crashreportcopymobile")
                break
            except:
                import traceback
                traceback.print_exc()
                retry -= 1
                time.sleep(5)
        if self.service is None:
            raise RuntimeError('Connect to crashreportcopymobile failed')
        self.afc_shell = AFCShell2(afcclient=self)
        super(AFCCrashLog, self).__init__(self.lockdown, service=self.service)
    
    def _filter_crash_log(self, logs, log_path, procname):
        latest_log = None
        latest_timestamp = 0
        for log in logs:
            result = re.match('%s/%s.+(\d{4}-\d{2}-\d{2}-\d{6}).*\.(ips|synced)' % (log_path, procname), log)
            if result:
                time_src = result.group(1)
                timestamp = int(time.mktime(time.strptime(time_src, "%Y-%m-%d-%H%M%S")))
                if timestamp > latest_timestamp:
                    latest_log = log
                    latest_timestamp = timestamp
        
        # 过滤掉Xcode的bug引起的假crash
        now = int(time.time()) 
        if now - latest_timestamp > 300: #判断crash文件的时间戳距离当前时间超过5分钟，则认为当前没有发生crash
            latest_log = None
        return latest_log
    
    def get_crash_log(self, procname, dest, is_delete = False):
        local_crashes =[]
        remote_crash_path = '/'
        for filename in self.afc_shell.dir_walk(remote_crash_path):
            if procname in filename:
                remote_crash_file = os.path.join(remote_crash_path, filename)
                data = self.afc_shell.get_file_contents(remote_crash_file)
                local_crash_file = os.path.join(dest, filename)
                local_crashes.append(local_crash_file)
                with open(local_crash_file, 'wb') as fp:
                    fp.write(data)
                if is_delete: 
                    self.afc_shell.do_rm(remote_crash_file)
        return self._filter_crash_log(local_crashes, dest, procname)
    
    def close(self):
        if self.service:
            self.service.close()


class SandboxClient(AFCClient):
    '''
    访问app的sandbox的类
    '''

    def __init__(self, udid=None, bid='com.tencent.sng.test.gn', sandbox="VendContainer", lockdown=None):
        self.lockdown = lockdown if lockdown else LockdownClient(udid)
        self.bid = bid
        self.service = None
        self.udid = udid
        retry = 5
        while retry > 0:
            try:
                self.service = self.lockdown.startService("com.apple.mobile.house_arrest")
                break
            except:
                import traceback
                traceback.print_exc()
                retry -= 1
                time.sleep(2)
                if retry <= 0:
                    raise
        if self.service is None:
            raise RuntimeError('Connect to house_arrest failed')

        self.service.sendPlist({"Command": sandbox, "Identifier": bid})
        status = self.service.recvPlist()
        if 'Error' in status and status['Error'] == "ApplicationLookupFailed":
            raise RuntimeWarning('ApplicationLookupFailed')
        if 'Status' in status and status['Status'] != 'Complete':
            raise RuntimeWarning('House arrest service launch failed')
        super(SandboxClient, self).__init__(self.lockdown, service=self.service)

        self.afc_shell = AFCShell2(self)
    
    def copy_to_local(self, remotepath, localpath='/tmp', is_dir=False, is_delete = False):
        '''拷贝手机中sandbox指定目录的文件到Mac本地
        :param remotepath: sandbox上的目录或者文件，例如：/Library/Caches/test/
        :type remotepath: str        
        :param localpath: 本地的目录
        :type localpath: str
        :param is_dir: remotepath是否为目录，默认为单个文件
        :type is_dir: bool
        :param is_delete: 是否删除，默认为单个文件
        :type is_delete: bool
        :return: list   拷贝后的本地文件列表
        '''
        local_files = []
        if is_dir:
            remotefiles = self.afc_shell.dir_walk(remotepath)
        else:
            filepath, filename = os.path.split(remotepath)
            remotefiles = [filename]
            remotepath = filepath
        
        for f in remotefiles:
            src = os.path.join(remotepath,f)
            content = self.afc_shell.get_file_contents(src)
            if self.udid:
                (short_name, extension) = os.path.splitext(f)
                new_f = '%s-%s%s' % (short_name, self.udid, extension)
            if content is None:
                continue
            dst = os.path.join(localpath,new_f)
            local_files.append(dst)
            with open(dst, 'wb') as fd:
                fd.write(content)
            if is_delete:
                self.afc_shell.do_rm(src)
        return local_files

    def copy_to_remote(self, localfile, remotepath):
        '''拷贝Mac本地文件到手机中sandbox的指定目录
        :param localfile: Mac本地文件路径（仅限文件拷贝）
        :type localfile: str
        :param remotepath: sandbox上的目录，例如：/Library/Caches/
        :type remotepath: str
        :return: bool
        ''' 
        result = False
        with open(localfile, 'rb') as fd:
            data = fd.read()
            remotefile = os.path.join(remotepath,os.path.basename(localfile))
            status = self.afc_shell.set_file_contents(remotefile,data)
            if status == 0:
                result = True
        return result
    
    def remove(self, files):
        '''删除sandbox中的文件或者目录
        '''
        self.afc_shell.do_rm(files)
    
    def close(self):
        '''必须显示调用该函数，否则会出现端口泄漏
        '''
        if self.service:
            self.service.close()


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''QT4i egg命令行接口(仅供UISpy使用)
'''


import sys
import os


proj_root = os.path.dirname(os.path.abspath(__file__))
pythonpaths = []
srcpypath = os.environ.get('PYTHONPATH', '')
exlib_dir = os.path.dirname(proj_root)
for filename in os.listdir(exlib_dir):
    if filename.endswith('.egg'):
        lib_path = os.path.join(exlib_dir, filename)
        if os.path.isfile(lib_path) and lib_path not in sys.path:
            sys.path.insert(0, lib_path)
            pythonpaths.append(lib_path)

os.environ['PYTHONPATH'] = srcpypath + ':'.join(pythonpaths)


if __name__ == '__main__':
    from testbase.management import ManagementTools
    ManagementTools().run()


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''WebDriver命令
'''

class Command(object):
    """
    Defines constants for the standard WebDriver commands.

    While these constants have no meaning in and of themselves, they are
    used to marshal commands through a service that implements WebDriver's
    remote wire protocol:

        https://github.com/SeleniumHQ/selenium/wiki/JsonWireProtocol

    """

    # Keep in sync with org.openqa.selenium.remote.DriverCommand

    HEALTH = "health"
    DEACTIVATE_APP = "deactivateApp"
    STATUS = "status"
    SESSION = "Session"
    GET_ALL_SESSIONS = "getAllSessions"
    DELETE_SESSION = "deleteSession"
    CLOSE = "close"
    START = "start"
    QUIT = "quit"
    GET = "get"
    FIND_ELEMENT = "findElement"
    FIND_ELEMENTS = "findElements"
    FIND_CHILD_ELEMENT = "findChildElement"
    FIND_CHILD_ELEMENTS = "findChildElements"
    CLEAR_ELEMENT = "clearElement"
    CLICK_ELEMENT = "clickElement"
    SEND_KEYS_TO_ELEMENT = "sendKeysToElement"
    SEND_KEYS_TO_ACTIVE_ELEMENT = "sendKeysToActiveElement"
    SUBMIT_ELEMENT = "submitElement"
    GET_WINDOW_SIZE = "getWindowSize"
    GET_WINDOW_POSITION = "getWindowPosition"
    SET_WINDOW_SIZE = "setWindowSize"
    SET_WINDOW_POSITION = "setWindowPosition"
    SWITCH_TO_WINDOW = "switchToWindow"
    SWITCH_TO_FRAME = "switchToFrame"
    SWITCH_TO_PARENT_FRAME = "switchToParentFrame"
    GET_ACTIVE_ELEMENT = "getActiveElement"
    GET_CURRENT_URL = "getCurrentUrl"
    GET_PAGE_SOURCE = "getPageSource"
    GET_TITLE = "getTitle"
    EXECUTE_SCRIPT = "executeScript"
    GET_ELEMENT_TREE = "getElementTree"
    GET_ELEMENT_TEXT = "getElementText"
    GET_ELEMENT_VALUE = "getElementValue"
    GET_ELEMENT_TAG_NAME = "getElementTagName"
    SET_ELEMENT_SELECTED = "setElementSelected"
    IS_ELEMENT_SELECTED = "isElementSelected"
    IS_ELEMENT_ENABLED = "isElementEnabled"
    IS_ELEMENT_DISPLAYED = "isElementDisplayed"
    GET_ELEMENT_LOCATION = "getElementLocation"
    GET_ELEMENT_LOCATION_ONCE_SCROLLED_INTO_VIEW = "getElementLocationOnceScrolledIntoView"
    GET_ELEMENT_SIZE = "getElementSize"
    GET_ELEMENT_RECT = "getElementRect"
    GET_ELEMENT_ATTRIBUTE = "getElementAttribute"
    ELEMENT_EQUALS = "elementEquals"
    SCREENSHOT = "screenshot"
    ELEMENT_SCREENSHOT = "elementScreenshot"
    IMPLICIT_WAIT = "implicitlyWait"
    EXECUTE_ASYNC_SCRIPT = "executeAsyncScript"
    SET_SCRIPT_TIMEOUT = "setScriptTimeout"
    SET_TIMEOUTS = "setTimeouts"
    MAXIMIZE_WINDOW = "windowMaximize"

    #Alerts
    DISMISS_ALERT = "dismissAlert"
    ACCEPT_ALERT = "acceptAlert"
    SET_ALERT_VALUE = "setAlertValue"
    GET_ALERT_TEXT = "getAlertText"
    SET_ALERT_CREDENTIALS = "setAlertCredentials"

    # Advanced user interactions
    CLICK = "mouseClick"
    DOUBLE_CLICK = "mouseDoubleClick"
    MOUSE_DOWN = "mouseButtonDown"
    MOUSE_UP = "mouseButtonUp"
    MOVE_TO = "mouseMoveTo"

    # Screen Orientation
    SET_SCREEN_ORIENTATION = "setScreenOrientation"
    GET_SCREEN_ORIENTATION = "getScreenOrientation"

    # Touch Actions
    SINGLE_TAP = "touchSingleTap"
    TOUCH_DOWN = "touchDown"
    TOUCH_UP = "touchUp"
    TOUCH_MOVE = "touchMove"
    TOUCH_SCROLL = "touchScroll"
    DOUBLE_TAP = "touchDoubleTap"
    LONG_PRESS = "touchLongPress"
    FLICK = "touchFlick"
    FORCE_TOUCH = 'forceTouch'

    #HTML 5
    EXECUTE_SQL = "executeSql"

    GET_LOCATION = "getLocation"
    SET_LOCATION = "setLocation"

    GET_APP_CACHE = "getAppCache"
    GET_APP_CACHE_STATUS = "getAppCacheStatus"
    CLEAR_APP_CACHE = "clearAppCache"

    GET_LOCAL_STORAGE_ITEM = "getLocalStorageItem"
    REMOVE_LOCAL_STORAGE_ITEM = "removeLocalStorageItem"
    GET_LOCAL_STORAGE_KEYS = "getLocalStorageKeys"
    SET_LOCAL_STORAGE_ITEM = "setLocalStorageItem"
    CLEAR_LOCAL_STORAGE = "clearLocalStorage"
    GET_LOCAL_STORAGE_SIZE = "getLocalStorageSize"

    GET_SESSION_STORAGE_ITEM = "getSessionStorageItem"
    REMOVE_SESSION_STORAGE_ITEM = "removeSessionStorageItem"
    GET_SESSION_STORAGE_KEYS = "getSessionStorageKeys"
    SET_SESSION_STORAGE_ITEM = "setSessionStorageItem"
    CLEAR_SESSION_STORAGE = "clearSessionStorage"
    GET_SESSION_STORAGE_SIZE = "getSessionStorageSize"

    # Mobile
    GET_NETWORK_CONNECTION = "getNetworkConnection"
    SET_NETWORK_CONNECTION = "setNetworkConnection"
    CURRENT_CONTEXT_HANDLE = "getCurrentContextHandle"
    CONTEXT_HANDLES = "getContextHandles"
    SWITCH_TO_CONTEXT = "switchToContext"
    
    #QT4i
    QTA_DEVICE_CLICK = "qtaDeviceClick"
    QTA_DEVICE_DOUBLE_CLICK = "qtaDeviceDoubleClick"  
    QTA_DEVICE_LONG_CLICK = "qtaDeviceLongClick"
    QTA_DEVICE_DRAG = "qtaDeviceDrag"
    QTA_DEVICE_PINCH = "qtaDevicePinch"
    QTA_DEVICE_ROTATE = "qtaDeviceRotate"
    QTA_DEVICE_SENDKEYS = "qtaDeviceSendKeys"
    QTA_FIND_ELEMENTS =  "qtaFindElements"
    QTA_ELEMENT_CLICK = "qtaElementClick"
    QTA_ELEMENT_DOUBLE_CLICK = "qtaElementDoubleClick"
    QTA_ELEMENT_LONG_CLICK = "qtaElementLongClick"
    QTA_ELEMENT_DRAG = "qtaElementDrag"
    QTA_ELEMENT_SENDKEYS = "qtaElementSendKeys"
    QTA_GET_PARENT_ELEMENT = "qtaGetParent"
    QTA_GET_CHILDREN_ELEMENTS = "qtaGetChildren"
    QTA_SCROLL_TO_VISIBLE = "qtaScrollToVisible"
    QTA_DRAG_TO_VALUE = "qtaDragToValue"
    QTA_GET_ELEMENT_ATTRS = "qtaGetElementAttrs"
    QTA_ALERT_RULES_UPDATE = "qtaAlertRulesUpdate"
    QTA_ELEMENT_SETVALUE = "qtaElementSetValue"
    QTA_ELEMENT_TREE = "qtaElementTree"
    QTA_STOP_AGENT = "qtaStopAgent"
    QTA_GET_FOREGROUND_APP_NAME = "qtaGetForegroundAppName"
    QTA_GET_FOREGROUND_APP_PID = "qtaGetForegroundAppPid"
    QTA_GET_ELEMENT_WIN_INFO = "qtaGetElementWinInfo"
    QTA_ALERT_DISMISS = "qtaAlertDismiss"
    QTA_DEVICE_LOCK = "qtaDeviceLock"
    QTA_DEVICE_UNLOCK = "qtaDeviceUnlock"
    QTA_SANDBOX_LIST = "qtaSandboxList"
    QTA_SANDBOX_REMOVE = "qtaSandboxRemove"
    QTA_ALBUM_UPLOAD = "qtaAlbumUpload"
    QTA_STUB_CALL = "qtaStubCall"
    QTA_DEVICE_VOLUME = "qtaDeviceVolume"
    QTA_DEVICE_SIRI = "qtaDeviceSiri"
    QTA_DEVICE_SCREEN_DIRECTION = "qtaDeviceScreenDirection"
    QTA_DEVICE_DETAIL_INFO = "qtaDeviceDetailInfo"
    QTA_WHEEL_SELECT = "qtaWheelSelect"